# -*- coding: utf-8 -*-
"""
Created on Sun Dec 15 19:22:32 2013

@author: mel
"""
import wx
from math import sqrt
from ctx import theContext as context
import model
import app.resources as rc
import tran
import copy

NO_WHERE = 0
IN_CLASS = 1
IN_INHERIT = 2
IN_AGGREG = 3
IN_NOTE = 4
IN_LINE = 5
IN_LINE_MARK = 7
IN_SHAPE = 6


def RectCorners(r):
    """Obtiene los vertices en orden SE NE NW SW"""
    (x, y, w, h) = r.Get()
    p0 = wx.Point(x, y)
    p1 = wx.Point(x, y + h)
    p2 = wx.Point(x + w, y + h)
    p3 = wx.Point(x + w, y)
    return (p0, p1, p2, p3)


def RectSides(r):
    """Obtiene los lados de un rectangulo en orden RLBT"""
    (p0, p1, p2, p3) = RectCorners(r)
    return ([p2, p3], [p0, p1], [p1, p2], [p0, p3])


def RectMiddles(r):
    """Obtiene los puntos centrales de cara de un rectangulo en orden RLBT"""
    return [wx.Point((a.x + b.x) // 2,
        (a.y + b.y) // 2) for [a, b] in RectSides(r)]


def dot2(v1, v2):
    """Obtiene el producto escalar de dos vectores"""
    return (float(v1.x) * float(v2.x)) + (float(v1.y) * float(v2.y))


def norm(v):
    """obtiene la norma de un vector"""
    return sqrt(dot2(v, v))


def vfact(v, a):
    """multiplica un vector por un escalar"""
    x = float(v.x) * a
    y = float(v.y) * a
    return wx.Point(int(x), int(y))


def NearPoint(r, p):
    """Obtiene el punto mas proximo a p en el borde del rectangulo r
    y el indice de cara en orden RLBT"""
    c = RectSides(r)
    # print "RectSides:\n"
    # for side in c:
    #    print("[({x0},{y0}),({x1},{y1})]\n".format(
    #        x0=str(side[0].x), y0=str(side[0].y),
    #        x1=str(side[1].x), y1=str(side[1].y)))
    d2 = None
    ir = None
    m2 = None
    for i in range(4):
        segment = c[i]
        # los puntos de la recta son de la forma q(a) = c[0] + a*(c[1]-c[0])
        # el punto mas proximo es aquel para que (p-q(a))*(c[1]-c[0]) = 0, de
        # donde
        # (p-c[0])*(c[1]-c[0]) = a*||c[0]-c[1]||²  Si a esta en el rango
        # [0, 1], la solucion esta en [c[0],c[1]]. En caso contrario, el
        # extremo mas proximo será c[0] si a<0 o c[1] si a>1
        director = segment[1] - segment[0]
        m = dot2(director, director)
        a = dot2(p - segment[0], director) / m
        if a <= 0:
            u = p - segment[0]
            s = dot2(u, u)
            if d2 is None or d2 > s or (d2 == s and m < m2):
                # si la distancia es la primera o es menor que la
                # anteriormente determinada o es igual, pero el lado es
                # mas pequeño ...
                d2 = s
                ir = i
                m2 = m
                point = segment[0]
            continue
        elif a >= 1:
            u = p - segment[1]
            s = dot2(u, u)
            if d2 is None or d2 > s or (d2 == s and m < m2):
                # si la distancia es la primera o es menor que la
                # anteriormente determinada o es igual, pero el lado es
                # mas pequeño ...
                d2 = s
                ir = i
                m2 = m
                point = segment[1]
            continue
        else:
            q = segment[0] + vfact(director, a)
            u = p - q
            s = dot2(u, u)
            if d2 is None or d2 > s or (d2 == s and m < m2):
                # si la distancia es la primera o es menor que la
                # anteriormente determinada o es igual, pero el lado es
                # mas pequeño ...
                d2 = s
                ir = i
                m2 = m
                point = q
            continue
    return ir, point


class DiagramElement(object):
    """Base class for diagram elements"""
    def __init__(self, obj, pos):
        """Initialization"""
        super(DiagramElement, self).__init__()
        self._obj = obj
        self._pos = wx.Point(0, 0)
        if pos is not None:
            self._pos = wx.Point(pos.x, pos.y)
        self._selected = False

    @property
    def object(self):
        """Return the referent object"""
        return self._obj

    def Draw(self, ctx):
        """Draw element"""
        pass

    def HitTest(self, pos):
        """Check about element near pos"""
        return (None, NO_WHERE)

    def Select(self, value=True):
        """Set selected state"""
        self._selected = value

    def Selected(self):
        """Get selected state"""
        return self._selected

    def ToggleSelected(self):
        """Toggle selected state"""
        self._selected = not self._selected

    def Erase(self, dc):
        """Delete element"""
        pass

    def Remove(self, diag):
        """Remove element from class diagram"""
        pass

    def Shift(self, shift):
        """Displaces elements"""
        self._pos += shift


class ClosedShape(DiagramElement):
    """Represents a closed shape"""
    def __init__(self, obj, pos):
        """Initialization"""
        super(ClosedShape, self).__init__(obj, pos)
        self.Layout()
        self._rgn = self.Region()

    def __getstate__(self):
        """Get pickle context"""
        state = dict(self.__dict__)
        if '_rgn' in state:
            del state['_rgn']
        return state

    def __setstate__(self, state):
        """Set pickle context"""
        for attr in state:
            setattr(self, attr, state[attr])
        self._rgn = self.Region()

    def Layout(self):
        """Initialize geometrical properties"""
        pass

    def Shift(self, shift):
        """Displaces element"""
        super(ClosedShape, self).Shift(shift)
        self._rgn.Offset(shift.x, shift.y)

    def HitTest(self, pos):
        """Check about element near pos"""
        if self._rgn.ContainsPoint(pos) == wx.InRegion:
            return (self, IN_SHAPE)
        else:
            return (None, NO_WHERE)


class RectShape(ClosedShape):
    """Represents a rect, closed shape"""
    def __init__(self, obj, pos, **karg):
        """"""
        super(RectShape, self).__init__(obj, pos)
        if 'size' in karg:
            self._rectSize = karg['size']
        else:
            self._rectSize = None
        self.Layout()

    def Rect(self):
        """"""
        return wx.Rect(self._pos.x, self._pos.y,
            self._rectSize.x, self._rectSize.y)

    def RectSize(self):
        """Get rectangle"""
        return self._rectSize

    def Region(self):
        """Gets the class region"""
        return wx.Region(self._pos.x, self._pos.y,
            self._rectSize.x, self._rectSize.y)

    def HitTest(self, pos):
        """Check about element near pos"""
        r = wx.RectPS(self._pos, self._rectSize)
        if r.Contains(pos):
            return (self, IN_SHAPE)
        return (None, NO_WHERE)


class NoteElement(RectShape):
    """Represents a text note in diagram"""
    def __init__(self, obj, pos):
        """Initialization"""
        # in order to ensure correct class initialization we need to
        # first define the members, prior to super call
        super(NoteElement, self).__init__(obj, pos)

    def Layout(self):
        """Update some properties"""
        self._rectSize = wx.Size(0, 0)
        dc = wx.ScreenDC()
        (width, height) = dc.GetTextExtent(self._obj.note)
        self._textSize = wx.Size(width, height)
        rectSize = wx.Size(width + 45 - (width % 5), height + 25 - (height % 5))
        if not hasattr(self, '_rectSize'):
            self._rectSize = rectSize
        else:
            if rectSize.x > self._rectSize.x:
                self._rectSize.x = rectSize.x
            if rectSize.y > self._rectSize.y:
                self._rectSize.y = rectSize.y
        self._xOffset = (self._rectSize.x - width) // 2
        self._yOffset = (self._rectSize.y - height) // 2

    def Draw(self, dc):
        """Draw class element"""
        dc.SetLogicalFunction(wx.COPY)
        pen = wx.GREY_PEN
        if self.Selected():
            brush = wx.Brush(wx.TheColourDatabase.Find('GOLDENROD'))
        else:
            brush = wx.Brush(wx.TheColourDatabase.Find('THISTLE'))
        dc.SetPen(pen)
        dc.SetBrush(brush)
        if self.Selected():
            dc.DrawRectanglePointSize(self._pos, self._rectSize)
        else:
            dc.DrawRectanglePointSize(self._pos, self._rectSize)
        dc.DrawText(self._obj.note, self._pos.x +
            self._xOffset, self._pos.y + self._yOffset)

    def Erase(self, dc):
        """Erase element"""
        dc.SetLogicalFunction(wx.COPY)
        dc.SetPen(wx.WHITE_PEN)
        dc.SetBrush(wx.WHITE_BRUSH)
        dc.DrawRectanglePointSize(self._pos, self._rectSize)

    def Remove(self, diag):
        """Remove element from class diagram"""
        pass


class ClassElement(RectShape):
    """Represents a class in diagram"""
    def __init__(self, obj, pos):
        """Initialization"""
        # in order to ensure correct class initialization we need to
        # first define the members, prior to super call
        self._ancestors = []
        self._derivatives = []
        self._fromRelation = []
        self._toRelation = []
        super(ClassElement, self).__init__(obj, pos)

    @property
    def label(self):
        """Return the label of the element"""
        return self._obj.label

    def Layout(self):
        """Update some properties"""
        self._rectSize = wx.Size(0, 0)
        dc = wx.ScreenDC()
        (width, height) = dc.GetTextExtent(self.label)
        self._textSize = wx.Size(width, height)
        rectSize = wx.Size(width + 45 - (width % 5), height + 25 - (height % 5))
        sizeChange = False
        if not hasattr(self, '_rectSize'):
            self._rectSize = rectSize
            sizeChange = True
        else:
            if rectSize.x > self._rectSize.x:
                self._rectSize.x = rectSize.x
                sizeChange = True
            if rectSize.y > self._rectSize.y:
                self._rectSize.y = rectSize.y
                sizeChange = True
        self._xOffset = (self._rectSize.x - width) // 2
        self._yOffset = (self._rectSize.y - height) // 2
        if sizeChange:
            # se han de redimensionar las relaciones
            for inh in self._ancestors:
                inh.Initialize()
            for inh in self._derivatives:
                inh.Initialize()
            for rel in self._fromRelation:
                rel.Initialize()
            for rel in self._toRelation:
                rel.Initialize()

    def Draw(self, dc):
        """Draw class element"""
        dc.SetLogicalFunction(wx.COPY)
        pen = wx.GREY_PEN
        if self.Selected():
            brush = wx.Brush(wx.TheColourDatabase.Find('GOLDENROD'))
        else:
            brush = wx.Brush(wx.TheColourDatabase.Find('AQUAMARINE'))
        dc.SetPen(pen)
        dc.SetBrush(brush)
        if self.Selected():
            dc.DrawRectanglePointSize(self._pos, self._rectSize)
        else:
            dc.DrawRectanglePointSize(self._pos, self._rectSize)
        dc.DrawText(self.label, self._pos.x +
            self._xOffset, self._pos.y + self._yOffset)

    def Erase(self, dc):
        """Erase element"""
        dc.SetLogicalFunction(wx.COPY)
        dc.SetPen(wx.WHITE_PEN)
        dc.SetBrush(wx.WHITE_BRUSH)
        dc.DrawRectanglePointSize(self._pos, self._rectSize)
        for e in self._ancestors:
            e.Erase(dc)
        for e in self._derivatives:
            e.Erase(dc)

    def Remove(self, diag):
        """Remove element from class diagram"""
        k = copy.copy(self._ancestors)
        for inh in k:
            diag.RemoveElement(inh)
        k = copy.copy(self._derivatives)
        for inh in k:
            diag.RemoveElement(inh)
        k = copy.copy(self._fromRelation)
        for rel in k:
            diag.RemoveElement(rel)
        k = copy.copy(self._fromRelation)
        for rel in k:
            diag.RemoveElement(rel)


class ConnectorSegment(object):
    """Represents a segment"""
    def __init__(self, start, end):
        """Initialization"""
        self._start = wx.Point(start.x, start.y)
        self._end = wx.Point(end.x, end.y)
        if self._start.x == self._end.x:
            self._orientation = wx.VERTICAL
        else:
            self._orientation = wx.HORIZONTAL
        self._mark = start + end
        self._mark.x //= 2
        self._mark.y //= 2
        self._mark.x -= 3
        self._mark.y -= 3

    def HitTest(self, pos, selected):
        """Check about mouse position"""
        if selected:
            # when selected, markers must selected first
            r = wx.Rect(self._mark.x, self._mark.y, 7, 7)
            if r.Contains(pos):
                return (self, IN_LINE_MARK)
        if self._orientation == wx.HORIZONTAL:
            if (self._start.x - pos.x) * (self._end.x - pos.x) < 0:
                if abs(self._start.y - pos.y) < 4:
                    return (self, IN_LINE)
        else:
            if (self._start.y - pos.y) * (self._end.y - pos.y) < 0:
                if abs(self._start.x - pos.x) < 4:
                    return (self, IN_LINE)
        return (None, NO_WHERE)

    def Draw(self, dc):
        """Draw element"""
        dc.DrawLine(self._start.x, self._start.y, self._end.x, self._end.y)

    def DrawMark(self, dc):
        """Draw selected marks"""
        sz = wx.Size(7, 7)
        dc.DrawRectanglePointSize(self._mark, sz)

    def Shift(self, shift):
        """Shift position"""
        self._start += shift
        self._end += shift
        self._mark += shift


class ConnectorStartSegment(ConnectorSegment):
    """Represents an start segment"""
    def __init__(self, start, end):
        """Initialization"""
        super(ConnectorStartSegment, self).__init__(start, end)
        if self._orientation is wx.HORIZONTAL:
            self._mark.x = self._start.x - 3
        else:
            self._mark.y = self._start.y - 3
        self._touch = False


class ArrowStartSegment(ConnectorStartSegment):
    """Represents an start arrow"""
    def __init__(self, start, end):
        """Initialization"""
        super(ArrowStartSegment, self).__init__(start, end)

    def Draw(self, dc):
        """Draw element"""
        p = self._start
        if self._orientation is wx.HORIZONTAL:
            if p.x > self._end.x:
                q = wx.Point(p.x - 10, p.y - 5)
                r = wx.Point(p.x - 10, p.y + 5)
            else:
                q = wx.Point(p.x + 10, p.y - 5)
                r = wx.Point(p.x + 10, p.y + 5)
        elif self._orientation is wx.VERTICAL:
            if p.y > self._end.y:
                q = wx.Point(p.x - 5, p.y - 10)
                r = wx.Point(p.x + 5, p.y - 10)
            else:
                q = wx.Point(p.x - 5, p.y + 10)
                r = wx.Point(p.x + 5, p.y + 10)
        else:
            return
        self._points = [p, q, p, r, p, self._end]
        dc.DrawLines(self._points)


class LanceStartSegment(ConnectorStartSegment):
    """Represents an lance"""
    def __init__(self, start, end):
        """Initialization"""
        super(LanceStartSegment, self).__init__(start, end)

    def Draw(self, dc):
        """Draw element"""
        p = self._start
        if self._orientation is wx.HORIZONTAL:
            if p.x > self._end.x:
                q = wx.Point(p.x - 10, p.y - 5)
                r = wx.Point(p.x - 10, p.y + 5)
                s = wx.Point(p.x - 10, p.y)
            else:
                q = wx.Point(p.x + 10, p.y - 5)
                r = wx.Point(p.x + 10, p.y + 5)
                s = wx.Point(p.x + 10, p.y)
        elif self._orientation is wx.VERTICAL:
            if p.y > self._end.y:
                q = wx.Point(p.x - 5, p.y - 10)
                r = wx.Point(p.x + 5, p.y - 10)
                s = wx.Point(p.x, p.y - 10)
            else:
                q = wx.Point(p.x - 5, p.y + 10)
                r = wx.Point(p.x + 5, p.y + 10)
                s = wx.Point(p.x, p.y + 10)
        else:
            return
        self._points = [p, q, s, self._end, s, r, p]
        dc.DrawLines(self._points)


class DiamondStartSegment(ConnectorStartSegment):
    """Represents an diamod <>"""
    def __init__(self, start, end):
        """Initialization"""
        super(DiamondStartSegment, self).__init__(start, end)

    def Draw(self, dc):
        """Draw element"""
        p = self._start
        if self._orientation is wx.HORIZONTAL:
            if p.x > self._end.x:
                q = wx.Point(p.x - 10, p.y - 5)
                r = wx.Point(p.x - 10, p.y + 5)
                s = wx.Point(p.x - 20, p.y)
            else:
                q = wx.Point(p.x + 10, p.y - 5)
                r = wx.Point(p.x + 10, p.y + 5)
                s = wx.Point(p.x + 20, p.y)
        elif self._orientation is wx.VERTICAL:
            if p.y > self._end.y:
                q = wx.Point(p.x - 5, p.y - 10)
                r = wx.Point(p.x + 5, p.y - 10)
                s = wx.Point(p.x, p.y - 20)
            else:
                q = wx.Point(p.x - 5, p.y + 10)
                r = wx.Point(p.x + 5, p.y + 10)
                s = wx.Point(p.x, p.y + 20)
        else:
            return
        self._points = [p, q, s, self._end, s, r, p]
        dc.DrawLines(self._points)


class ConnectorEndSegment(ConnectorSegment):
    """Represents an start segment"""
    def __init__(self, start, end):
        """Initialization"""
        super(ConnectorEndSegment, self).__init__(start, end)
        if self._orientation is wx.HORIZONTAL:
            self._mark.x = self._end.x - 3
        else:
            self._mark.y = self._end.y - 3
        self._touch = False



class ArrowEndSegment(ConnectorEndSegment):
    """Represents an end arrow"""
    def __init__(self, start, end):
        """Initialization"""
        super(ArrowEndSegment, self).__init__(start, end)

    def Draw(self, dc):
        """Draw element"""
        p = self._end
        if self._orientation is wx.HORIZONTAL:
            if p.x > self._start.x:
                q = wx.Point(p.x - 10, p.y - 5)
                r = wx.Point(p.x - 10, p.y + 5)
            else:
                q = wx.Point(p.x + 10, p.y - 5)
                r = wx.Point(p.x + 10, p.y + 5)
        elif self._orientation is wx.VERTICAL:
            if p.y > self._start.y:
                q = wx.Point(p.x - 5, p.y - 10)
                r = wx.Point(p.x + 5, p.y - 10)
            else:
                q = wx.Point(p.x - 5, p.y + 10)
                r = wx.Point(p.x + 5, p.y + 10)
        else:
            return
        self._points = [p, q, p, r, p, self._start]
        dc.DrawLines(self._points)


class LanceEndSegment(ConnectorEndSegment):
    """Represents an lance"""
    def __init__(self, start, end):
        """Initializes lance"""
        super(LanceEndSegment, self).__init__(start, end)

    def Draw(self, dc):
        """Draw element"""
        p = self._end
        if self._orientation is wx.HORIZONTAL:
            if p.x > self._start.x:
                q = wx.Point(p.x - 10, p.y - 5)
                r = wx.Point(p.x - 10, p.y + 5)
            else:
                q = wx.Point(p.x + 10, p.y - 5)
                r = wx.Point(p.x + 10, p.y + 5)
        elif self._orientation is wx.VERTICAL:
            if p.y > self._start.y:
                q = wx.Point(p.x - 5, p.y - 10)
                r = wx.Point(p.x + 5, p.y - 10)
            else:
                q = wx.Point(p.x - 5, p.y + 10)
                r = wx.Point(p.x + 5, p.y + 10)
        else:
            return
        self._points = [p, q, r, self._start, r, p]
        dc.DrawLines(self._points)


class DiamondEndSegment(ConnectorEndSegment):
    """Represents an diamod <>"""
    def __init__(self, start, end):
        """Initialization"""
        super(DiamondEndSegment, self).__init__(start, end)

    def Draw(self, dc):
        """Draw element"""
        p = self._end
        if self._orientation is wx.HORIZONTAL:
            if p.x > self._start.x:
                q = wx.Point(p.x - 10, p.y - 5)
                r = wx.Point(p.x - 10, p.y + 5)
                s = wx.Point(p.x - 20, p.y)
            else:
                q = wx.Point(p.x + 10, p.y - 5)
                r = wx.Point(p.x + 10, p.y + 5)
                s = wx.Point(p.x + 20, p.y)
        elif self._orientation is wx.VERTICAL:
            if p.y > self._end.y:
                q = wx.Point(p.x - 5, p.y - 10)
                r = wx.Point(p.x + 5, p.y - 10)
                s = wx.Point(p.x, p.y - 20)
            else:
                q = wx.Point(p.x - 5, p.y + 10)
                r = wx.Point(p.x + 5, p.y + 10)
                s = wx.Point(p.x, p.y + 20)
        else:
            return
        self._points = [p, q, s, self._start, s, r, p]
        dc.DrawLines(self._points)


class ConectorElement(DiagramElement):
    """Represents a conector between diagram elements"""
    def __init__(self, obj, fromShape, toShape):
        """Initialize a connector element"""
        super(ConectorElement, self).__init__(obj, None)
        self._FROM = fromShape
        self._TO = toShape
        self.segments = []

    def SelectFromClients(self):
        """Select element only if both extrems are selected"""
        self._selected = self._FROM._selected and self._TO._selected

    def DrawLines(self, dc):
        """Draw lines"""
        for segment in self.segments:
            segment.Draw(dc)

    def DrawMarkers(self, dc):
        """Draw track markers"""
        for segment in self.segments:
            segment.DrawMark(dc)

    def RefreshChanges(self):
        """Actualiza los extremos de los segmentos"""
        if len(self.segments) > 1:
            s = self.segments[0]
            e = self.segments[-1]
            self.segments[0] = self.StartSegment(s._start, s._end)
            self.segments[-1] = self.EndSegment(e._start, e._end)

    def StartSegment(self, p0, p1):
        """Create a new start segment"""
        return ConnectorStartSegment(p0, p1)

    def EndSegment(self, p0, p1):
        """Create a new end segment"""
        return ConnectorEndSegment(p0, p1)

    def LineStep(self, p0, p1, orientation=None, start=False, end=False):
        """Create a line step from p0 to p1"""
        sx = p1.x - p0.x
        sy = p1.y - p0.y
        if abs(sx) >= abs(sy) and orientation is not wx.VERTICAL:
            if sx == 0:
                return None
            q = wx.Point(p1.x, p0.y)
        else:
            q = wx.Point(p0.x, p1.y)
        if start is True:
            return self.StartSegment(p0, q)
        elif end is True:
            return self.EndSegment(p0, q)
        else:
            return ConnectorSegment(p0, q)

    def ShiftSegment(self, index, pos):
        """Desplaza el segmento seleccionado"""
        #Lo primero, es determinar de que segmento se trata
        #print("called shift with pos value ({x},{y})".format(
        #    x=str(pos.x), y=str(pos.y)))
        segment = self.segments[index]
        if index == 0:
            # se trata de mover el primer segmento. Para esto,
            # se determina primero el lado mas proximo
            side, point = NearPoint(self._FROM.Rect(), pos)
            #print("NearPoint devuelve cara {cara} posicion ({x},{y})".format(
            #    x=str(point.x), y=str(point.y), cara=str(side)))
            last = self.segments[-1]
            target = last._start
            self.segments = []
            # add first segment
            q = point
            u = side % 4
            if u < 2:
                #change is in x
                qq = wx.Point(q.x +
                (1 - 2 * u) * self.GetStartConnectorSize().x, q.y)
            else:
                #change is in y
                qq = wx.Point(q.x, q.y +
                    (5 - 2 * u) * self.GetStartConnectorSize().y)
            l = self.LineStep(q, qq, start=True)
            l._touch = True
            if l is not None:
                self.segments.append(l)
            q = qq
            while True:
                l = self.LineStep(q, target)
                if l is None:
                    break
                q = l._end
                self.segments.append(l)
            # add last segment
            self.segments.append(last)
            self.JoinExtremes()
            return 0
        elif index == len(self.segments) - 1:
            # ok, en este caso se trata, reciprocamente, del ultimo segmento
            side, point = NearPoint(self._TO.Rect(), pos)
            first = self.segments[0]
            target = first._end
            self.segments = []
            q = point
            u = side % 4
            if u < 2:
                #change is in x
                qq = wx.Point(q.x +
                (1 - 2 * u) * self.GetEndConnectorSize().x, q.y)
            else:
                #change is in y
                qq = wx.Point(q.x, q.y +
                    (5 - 2 * u) * self.GetEndConnectorSize().y)
            l = self.LineStep(qq, q, end=True)
            l._touch = True
            q = qq
            if l is not None:
                self.segments.append(l)
            while True:
                l = self.LineStep(q, target)
                if l is None:
                    break
                q = l._end
                l._end = l._start
                l._start = q
                self.segments.insert(0, l)
            # add last segment
            self.segments.insert(0, first)
            self.JoinExtremes()
            return len(self.segments) - 1
        else:
            #se trata de mover un segmento del medio. En este caso,
            #la simple orientacion determina que es lo que se mueve
            if segment._orientation is wx.HORIZONTAL:
                segment._start.y = pos.y
                segment._end.y = pos.y
                segment._mark.y = pos.y - 3
                self.segments[index - 1]._end.y = pos.y
                if index > 1:
                    self.segments[index - 1]._mark.y = (
                        self.segments[index - 1]._end.y +
                        self.segments[index - 1]._start.y) // 2
                self.segments[index + 1]._start.y = pos.y
                if index < len(self.segments) - 2:
                    self.segments[index + 1]._mark.y = (
                        self.segments[index + 1]._end.y +
                        self.segments[index + 1]._start.y) // 2
            else:
                segment._start.x = pos.x
                segment._end.x = pos.x
                segment._mark.x = pos.x - 3
                self.segments[index - 1]._end.x = pos.x
                if index > 1:
                    self.segments[index - 1]._mark.x = (
                        self.segments[index - 1]._end.x +
                        self.segments[index - 1]._start.x) // 2
                self.segments[index + 1]._start.x = pos.x
                if index < len(self.segments) - 2:
                    self.segments[index + 1]._mark.x = (
                        self.segments[index + 1]._end.x +
                        self.segments[index + 1]._start.x) // 2
            return index

    def GetEndConnectorSize(self):
        """Return the connector size"""
        return wx.Size(20, 20)

    def GetStartConnectorSize(self):
        """Return the connector size"""
        return wx.Size(20, 20)

    def GetSidePositions(self, p, size):
        """Obtiene los puntos centrales de los lados de un cuadrado
        centrado en el punto p"""
        p0 = wx.Point(p.x + size.x, p.y)
        p1 = wx.Point(p.x - size.x, p.y)
        p2 = wx.Point(p.x, p.y + size.y)
        p3 = wx.Point(p.x, p.y - size.y)
        return [p0, p1, p2, p3]

    def JoinExtremes(self):
        """Revisa los extremos, y elimina segmentos adyacentes que solo
        son prolongaciones"""
        l = len(self.segments) - 2
        if l > 0:
            if self.segments[0]._orientation == self.segments[1]._orientation:
                self.segments[0]._end = self.segments[1]._end
                self.segments.pop(1)
                l = l - 1
        if l > 0:
            z = l + 1
            if self.segments[l]._orientation == self.segments[z]._orientation:
                self.segments[z]._start = self.segments[l]._start
                self.segments.pop(l)

    def Initialize(self):
        """Create initial lines"""
        # decide sides
        sz = self.GetStartConnectorSize()
        starts = RectMiddles(self._FROM.Rect().Inflate(sz.x, sz.y))
        sz = self.GetEndConnectorSize()
        ends = RectMiddles(self._TO.Rect().Inflate(sz.x, sz.y))
        # find best distance
        d2m = None
        si = 0
        # if the start segment or the end segment already exist
        # and is touched, respect it
        stouch = False
        etouch = False
        if len(self.segments):
            if self.segments[0]._touch:
                start = self.segments[0]
                starts = [start._start] * 4
                stouch = True
            if self.segments[-1]._touch:
                end = self.segments[-1]
                ends = [end._end] * 4
                etouch = True
        for i in range(0, 16):
            p = starts[i % 4]
            q = ends[i // 4]
            d2x = p.x - q.x
            d2y = p.y - q.y
            d2 = (d2x * d2x) + (d2y * d2y)
            if d2m is None or d2 < d2m:
                d2m = d2
                bestFrom = p
                bestTo = q
                si = i
        self.segments = []
        # add first segment
        if stouch:
            l = start
            bestFrom = start._end
        else:
            q = bestFrom
            u = si % 4
            if u < 2:
                #change is in x
                q = wx.Point(q.x +
                    (2 * u - 1) * self.GetStartConnectorSize().x, q.y)
            else:
                #change is in y
                q = wx.Point(q.x, q.y +
                    (2 * u - 5) * self.GetStartConnectorSize().y)
            l = self.LineStep(q, bestFrom, start=True)
        if l is not None:
            self.segments.append(l)
        q = bestFrom
        if etouch:
            bestTo = end._start

        while True:
            l = self.LineStep(q, bestTo)
            if l is None:
                break
            q = l._end
            self.segments.append(l)
        # add last line
        if etouch:
            l = end
        else:
            q = bestTo
            u = si // 4

            if u < 2:
                #change is in x
                q = wx.Point(q.x +
                    (2 * u - 1) * self.GetEndConnectorSize().x, q.y)
            else:
                #change is in y
                q = wx.Point(q.x, q.y +
                    (2 * u - 5) * self.GetEndConnectorSize().y)
            l = self.LineStep(bestTo, q, end=True)
        if l is not None:
            self.segments.append(l)

        # After create the segments, the last step is check
        # for join extremes with adjacent segments
        self.JoinExtremes()

        # The connector position is only used as reference
        # when moving mores selected objects so this value
        # can change without special meaning. This method
        # seems a good place to reset it and ensure that
        # dont goes to infinity
        self._pos = wx.Point(0, 0)
        return True

    def SelectedPen(self):
        """Returns the selected pen"""
        return wx.RED_PEN

    def DefaultPen(self):
        """Returns the default pen for use when no selected"""
        return wx.BLACK_PEN

    def Draw(self, dc):
        """Draw class element"""
        dc.SetLogicalFunction(wx.COPY)
        if self._selected:
            pen = self.SelectedPen()
        else:
            pen = self.DefaultPen()
        dc.SetPen(pen)
        self.DrawLines(dc)
        if self._selected:
            self.DrawMarkers(dc)

    def Erase(self, dc):
        """Draw class element"""
        dc.SetLogicalFunction(wx.COPY)
        pen = wx.Pen(wx.WHITE)
        pen.SetWidth(2)
        dc.SetPen(pen)
        self.DrawLines(dc)
        self.DrawMarkers(dc)

    def HitTest(self, pos):
        """Check about element near pos"""
        for p in self.segments:
            (a, where) = p.HitTest(pos, self._selected)
            if where is not NO_WHERE:
                return ((self, p), where)
            #if p._orientation == wx.HORIZONTAL:
            #    if (p._start.x - pos.x) * (p._end.x - pos.x) < 0:
            #        if abs(p._start.y - pos.y) < 2:
            #            return (self, IN_LINE)
            #else:
            #    if (p._start.y - pos.y) * (p._end.y - pos.y) < 0:
            #        if abs(p._start.x - pos.x) < 2:
            #            return (self, IN_LINE)
        return (None, NO_WHERE)

    def Shift(self, shift):
        """Displaces elements"""
        super(ConectorElement, self).Shift(shift)
        for segment in self.segments:
            segment.Shift(shift)


class RelationElement(ConectorElement):
    """Represents a relationship in diagram"""
    def __init__(self, obj, fromShape, toShape):
        """Initialization"""
        super(RelationElement, self).__init__(obj, fromShape, toShape)
        self.Initialize()

    def EndSegment(self, p0, p1):
        """Get end segment"""
        return ArrowEndSegment(p0, p1)

    def StartSegment(self, p0, p1):
        """Get initial segment"""
        if self._obj._FROM._minCardinal is None:
            return super(RelationElement, self).StartSegment(p0, p1)
        if self._obj._FROM._minCardinal == 0:
            return super(RelationElement, self).StartSegment(p0, p1)
        return DiamondStartSegment(p0, p1)

    def GetStartConnectorSize(self):
        """Return the connector size"""
        return wx.Size(20, 20)

    def Remove(self, diag):
        """Remove element from diagram"""
        self._FROM._toRelation.remove(self)
        self._TO._fromRelation.remove(self)

    def Draw(self, dc):
        """Do draw"""
        super(RelationElement, self).Draw(dc)

    def SelectedPen(self):
        """return the selected pen"""
        pen = wx.Pen(wx.RED)
        if self._obj._global:
            pen.SetWidth(2)
        else:
            pen.SetWidth(1)
        return pen

    def DefaultPen(self):
        """return the default pen"""
        if self._obj._critical:
            pen = wx.Pen(wx.RED)
        else:
            pen = wx.Pen(wx.BLACK)
        if self._obj._global:
            pen.SetWidth(2)
        else:
            pen.SetWidth(1)
        return pen


class InheritanceElement(ConectorElement):
    """Represents a inheritance in diagram"""
    def __init__(self, obj, fromShape, toShape):
        """Initialization"""
        super(InheritanceElement, self).__init__(obj, fromShape, toShape)
        self.Initialize()

    def StartSegment(self, p0, p1):
        """Initialize segment"""
        return LanceStartSegment(p0, p1)

    def GetStartConnectorSize(self):
        """Return the connector size"""
        return wx.Size(20, 20)

    def Remove(self, diag):
        """Remove element from diagram"""
        self._FROM._derivatives.remove(self)
        self._TO._ancestors.remove(self)

    def Shift(self, shift):
        """Displaces elements"""
        super(InheritanceElement, self).Shift(shift)

    def DefaultPen(self):
        """Return default pen"""
        return wx.Pen(wx.BLUE)


class ClassDiagram(model.TComponent):
    """Implements a class diagram. This is the descriptor for the
    shapes and placements of the diagram."""

    @tran.TransactionalMethod('move class diagram {}')
    def drop(self, to):
        """Drops class diagram inside project, class or folder """
        target = to.inner_diagram_container
        if not target or self.project != target.project:
            return False  # avoid move classes between projects
        index = 0
        tran.TransactionalMoveObject(
            object=self, origin=self.parent, target=target, index=index)
        return True

    def __init__(self, **kwargs):
        """Initialization"""
        self._elements = kwargs.get('elements', [])
        self._zoom = kwargs.get('zoom', 1.0)
        # A4 300 dpi default size
        self._width = kwargs.get('width', 2480)
        self._height = kwargs.get('height', 3508)
        self._pane = None
        super(ClassDiagram, self).__init__(**kwargs)

    def get_kwargs(self):
        """Returns the kwargs needed for this object"""
        kwargs = {}
        kwargs['elements'] = self._elements
        kwargs['zoom'] = self._zoom
        kwargs['width'] = self._width
        kwargs['height'] = self._height
        kwargs.update(super(ClassDiagram, self).get_kwargs())
        return kwargs

    def AddClass(self, cls, pos):
        """Add class element"""
        nwelem = ClassElement(cls, pos)
        # add inheritance and derivative links
        self_cls = lambda x: x.inner_class == cls
        inherits = cls(model.cc.Inheritance, filter=self_cls, cut=True)
        for inh in inherits:
            elem = self.FindElement(inh._ancestor)
            if elem is not None:
                self.AddInheritance(inh, elem, nwelem)
        for elem in self._elements:
            if type(elem) is ClassElement:
                inherits = elem._obj(model.cc.Inheritance)
                for inh in inherits:
                    if inh._ancestor == cls:
                        self.AddInheritance(inh, nwelem, elem)
                        break
        # add relation links
        fromRelation = cls(model.cc.RelationFrom, filter=self_cls, cut=True)
        toRelation = cls(model.cc.RelationTo, filter=self_cls, cut=True)
        for rel in fromRelation:
            elem = self.FindElement(rel.GetFrom())
            if elem is not None:
                self.AddRelation(rel._key, elem, nwelem)
        for rel in toRelation:
            elem = self.FindElement(rel.GetTo())
            if elem is not None:
                self.AddRelation(rel._key, nwelem, elem)
        self._elements.append(nwelem)
        return nwelem

    def AddNote(self, obj, pos):
        """Add note element"""
        nwelem = NoteElement(obj, pos)
        self._elements.append(nwelem)
        return nwelem

    def AddInheritance(self, obj, fromShape, toShape):
        """Add inheritance element"""
        elem = InheritanceElement(obj, fromShape, toShape)
        fromShape._derivatives.append(elem)
        toShape._ancestors.append(elem)
        self._elements.append(elem)
        return elem

    def AddRelation(self, obj, fromShape, toShape):
        """Add relationship element"""
        elem = RelationElement(obj, fromShape, toShape)
        fromShape._toRelation.append(elem)
        toShape._fromRelation.append(elem)
        self._elements.append(elem)
        return elem

    def MoveElement(self, element, pos):
        """Move a element"""
        shift = pos - element._pos
        element._pos = pos

        if type(element) is ClassElement:
            for e in element._ancestors:
                e.Shift(shift)
                e.Initialize()
            for e in element._derivatives:
                e.Shift(shift)
                e.Initialize()
            for e in element._fromRelation:
                e.Shift(shift)
                e.Initialize()
            for e in element._toRelation:
                e.Shift(shift)
                e.Initialize()

    def BringToFront(self, element):
        """Move element to front"""
        if element in self._elements:
            self._elements.remove(element)
        self._elements.append(element)

    def RemoveElement(self, element):
        """Remove element from class diagram"""
        element.Remove(self)
        if element in self._elements:
            self._elements.remove(element)

    @property
    def selected_elements(self):
        """Return the list of selected elements"""
        return [x for x in self._elements if x._selected]

    def EraseSelected(self, dc):
        """Clear selected elements draws"""
        for element in self._elements:
            if element._selected:
                element.Erase(dc)
        # Erase inheritances
        for element in self._elements:
            if isinstance(element, ConectorElement):
                if not element._FROM._selected or not element._TO._selected:
                    element.Erase(dc)
                elif not element._selected:
                    element.Erase(dc)

    def ShiftSelected(self, shift):
        """Displace selected elements"""
        # Select inheritances with both extremes selected
        for element in self._elements:
            if isinstance(element, ConectorElement):
                element.SelectFromClients()
        #do shift
        for element in self._elements:
            if element._selected:
                element.Shift(shift)
                if type(element) is ClassElement:
                    for e in element._ancestors:
                        if not e._selected:
                            e.segments[-1].Shift(shift)
                            e.Initialize()
                    for e in element._derivatives:
                        if not e._selected:
                            e.segments[0].Shift(shift)
                            e.Initialize()
                    for e in element._fromRelation:
                        if not e._selected:
                            e.segments[-1].Shift(shift)
                            e.Initialize()
                    for e in element._toRelation:
                        if not e._selected:
                            e.segments[0].Shift(shift)
                            e.Initialize()
        # Update affected connectos
        # for element in self._elements:
        #    if isinstance(element, ConectorElement):
        #        if not element._selected:
        #            element.Initialize()

    def OnUndoRedoChanged(self):
        """Update from app"""
        if hasattr(self, '_pane'):
            # some bug happens here when we undo open project with open panes
            # and redo open
            from activity.models.ui import pane
            if type(self._pane) is not pane.DiagramPane:
                delattr(self, '_pane')
            else:
                book = context.frame.docBook
                index = book.GetPageIndex(self._pane)
                book.SetPageText(index, self.tab_label)
                book.SetPageBitmap(index, self.GetTabBitmap())
                self._pane.Refresh()
        #refresh relations
        for element in self._elements:
            if type(element) is InheritanceElement:
                felem = element._FROM
                telem = element._TO
                if felem not in self._elements or telem not in self._elements:
                    print('detected orphan inheritance in diagram')
                if element not in felem._derivatives:
                    felem._derivatives.append(element)
                if element not in telem._ancestors:
                    telem._ancestors.append(element)
            elif type(element) is RelationElement:
                felem = element._FROM
                telem = element._TO
                if felem not in self._elements or telem not in self._elements:
                    print('detected orphan relationship in diagram')
                if element not in felem._toRelation:
                    felem._toRelation.append(element)
                if element not in telem._fromRelation:
                    telem._fromRelation.append(element)
        super(ClassDiagram, self).OnUndoRedoChanged()

    def OnUndoRedoRemoving(self, root=True):
        """Prepare for delete"""
        if hasattr(self, '_pane') and not self._pane is None:
            book = context.frame.docBook
            page = self._pane
            delattr(self, '_pane')
            page.Commit()
            self._paneIndex = book.GetPageIndex(page)
            book.DeletePage(self._paneIndex)
        super(ClassDiagram, self).OnUndoRedoRemoving(root)

    def OnUndoRedoAdd(self):
        """Restore object from undo"""
        super(ClassDiagram, self).OnUndoRedoAdd()
        if hasattr(self, '_paneIndex'):
            from activity.models.ui import pane
            book = context.frame.docBook
            page = pane.DiagramPane(book, self)
            self._pane = page
            book.InsertPage(self._paneIndex, page, self.tab_label,
                False, self.bitmap_index)
            delattr(self, '_paneIndex')

    def GetTabBitmap(self):
        """Get the bitmap for tab control"""
        return rc.GetBitmap('classdiagram')
        #return GetBitmap(4)

    @property
    def bitmap_index(self):
        """Index of tree image"""
        return rc.GetBitmapIndex('classdiagram')

    def UnselectAll(self):
        """Unselect all elements"""
        for element in self._elements:
            element.Select(False)

    def DrawClassDiagram(self, ctx, skip=None):
        """Draw class diagram"""
        for element in self._elements:
            if element._obj == skip:
                continue
            element.Draw(ctx)

    def __getstate__(self):
        """Set picke context"""
        state = dict(self.__dict__)
        if '_parentIndex' in state:
            del state['_parentIndex']
        if '_pane' in state:
            del state['_pane']
        if '_paneIndex' in state:
            del state['_paneIndex']
        return state

    def PositionClass(self, cls, pos):
        """Reposition a class element"""
        cls.project.SetModified(True)
        for element in self._elements:
            if element._obj == cls:
                element._pos = pos
                return element
        element = ClassElement(cls, pos)
        self._elements.append(element)
        return element

    def HitTest(self, pos):
        """Check about element near pos"""
        result = (None, NO_WHERE)
        for element in self._elements:
            (a, b) = element.HitTest(pos)
            if b is not NO_WHERE:
                if element._selected:
                    return a, b
                else:
                    result = (a, b)
        return result

    def FindElement(self, obj):
        """Finds a class diagram representation for obj"""
        for x in self._elements:
            if x._obj == obj:
                return x
        return None



# -*- coding: utf-8 -*-
"""
Created on Sun Dec 15 19:22:32 2013

@author: mel
"""
import model
import app.resources as rc
from tran import TransactionStack


class Type(model.TComponent):
    """Implements class representation"""
    def __init__(self, **kwargs):
        """Initialization method"""
        self._access = kwargs.get('access', "public")
        self._definition = kwargs.get('definition', "")
        self._template = kwargs.get('template', None)
        super(Type, self).__init__(**kwargs)
        self.project.ExportCppCodeFiles(force=True)

    def Delete(self):
        """Handle delete"""
        project = self.project
        super(Type, self).Delete()
        project.ExportCppCodeFiles(force=True)

    def get_kwargs(self):
        """Returns the kwargs needed for this object"""
        kwargs = {}
        kwargs['access'] = self._access
        kwargs['definition'] = self._definition
        kwargs['template'] = self._template
        kwargs.update(super(Type, self).get_kwargs())
        return kwargs

    @property
    def template_args(self):
        """Retorna una cadena que representa los argumentos del patron.
        Por ejemplo, si la cadena del template es 'typename T,class S=x'
        retorna T,S. Por el momento no se considera el uso de variadic"""
        if self._template is None:
            return ''
        s = str(self.template_types)[1:-1]
        s = s.replace("u'", '')
        s = s.replace('u"', '')
        s = s.replace("'", '')
        s = s.replace('"', '')
        return s

    @property
    def template_types(self):
        """Retorna la lista de identificadores de tipo utilizados
        en la declaracion de patron"""
        if self._template is None:
            return []
        l = self._template.split(',')
        prologs = ['int', 'class', 'typename']
        r = []
        for u in l:
            v = u.split('=')[0]
            for p in prologs:
                v = v.replace(p, '')
            r.append(v.strip())
        return r

    def RemoveRelations(self):
        """Utility for undo/redo"""
        super(Type, self).RemoveRelations()

    def RestoreRelations(self):
        """Utility for undo/redo"""
        super(Type, self).RestoreRelations()

    def OnUndoRedoRemoving(self, root=True):
        """Prepare object to delete"""
        super(Type, self).OnUndoRedoRemoving(root)

    @property
    def can_delete(self):
        """Comprueba si el tipo puede ser eliminado"""
        if self._readOnly:
            return False
        project = self.project
        #Obtenemos la lista de miembros de clase
        if not project is None:
            #comprobamos si el tipo esta siendo utilizado
            typed = project(model.cc.MemberData, model.cc.MemberMethod, model.cc.Argument)
            for x in typed:
                if x._typei._type == self:
                    return False
            return True

    def OnUndoRedoChanged(self):
        """Update from app"""
        #Obtenemos la lista de miembros de clase
        super(Type, self).OnUndoRedoChanged()
        if not TransactionStack.InUndoRedo():
            k = self.outer_class or self.outer_module
            if k:
                k.ExportCppCodeFiles(force=True)

    def OnUndoRedoAdd(self):
        """Restore object from undo"""
        super(Type, self).OnUndoRedoAdd()

    @property
    def bitmap_index(self):
        """Index of tree image"""
        return rc.GetBitmapIndex("type", self._access)

    @property
    def scoped(self):
        """Get the scope"""
        return '{self.parent.scope}{self._name}'.format(self=self)

    def scoped_str(self, scope_list):
        """Gets the string representation with minimal relative scope"""
        s = self.scoped
        best_match = 0
        best_scope = None
        for x in scope_list:
            if s.find(x) == 0:
                if len(x) > best_match:
                    best_match = len(x)
                    best_scope = x
        if best_scope:
            s = s[best_match:]
            if s[:2] == '::':
                s = s[2:]
        return s

    @property
    def tree_label(self):
        """Get tree label"""
        # modify old versions
        if not hasattr(self, '_template'):
            self._template = None
        #endif
        if self._template is not None:
            return '{self._name}<{self.template_args}>'.format(self=self)
        return self._name


class typeinst(object):
    """This class represents the common usage of a type"""
    def __init__(self, **kwargs):
        """Initialice type instance"""
        self._type = kwargs.get('type', None)
        self._const = kwargs.get('const', False)
        self._ptr = kwargs.get('ptr', False)
        self._ref = kwargs.get('ref', False)
        self._ptr_to_ptr = kwargs.get('ptrptr', False)
        self._const_ptr = kwargs.get('constptr', False)
        self._array = kwargs.get('array', False)
        self._array_size = kwargs.get('arraysize', 0)
        self._funptr = kwargs.get('funptr', False)
        self._type_alias = kwargs.get('type_alias', None)   # for on-the-fly template types
        self._type_args = kwargs.get('type_args', None)   # for template types instance
        if self._funptr:
            self._argdecl = kwargs.get('argdecl', '()')
        super(typeinst, self).__init__()

    @property
    def type_name(self):
        """Return effective type name"""
        if self._type_alias is not None:
            return self._type_alias
        else:
            return self._type._name

    @property
    def is_ptr(self):
        """Return info about if the type is a pointer"""
        return self._ptr

    @property
    def base_type(self):
        """Return the base type"""
        return self._type

    @property
    def scoped(self):
        """Get the scope"""
        if self._type_alias is not None:
            return self._type_alias
        else:
            return self._type.scoped

    def scoped_str(self, scope_list):
        """Gets the string representation with minimal relative scope"""
        s = self.scoped
        best_match = 0
        best_scope = None
        for x in scope_list:
            if s.find(x) == 0:
                if len(x) > best_match:
                    best_match = len(x)
                    best_scope = x
        if best_scope:
            s = s[best_match:]
            if s[:2] == '::':
                s = s[2:]
        if self._type_args is not None:
            s = s + '<{}>'.format(self._type_args)
        if self._const:
            s = "const " + s
        if self._ptr:
            s += "*"
        if self._const_ptr:
            s += " const"
        if self._ptr_to_ptr:
            s += "*"
        if self._ref:
            s += "&"
        if self._funptr:
            s += ' (*{})' + self._argdecl
        else:
            s += ' {}'
        if self._array:
            if type(self._array_size) is str:
                s += '[{}]'.format(self._array_size)
            else:
                s += '[{}]'.format(str(self._array_size))
        return s

    def __str__(self):
        """Returns the type string format"""
        # migrate old versions
        if not hasattr(self, '_type_args'):
            self._type_args = None
        #end
        if self._type_alias is not None:
            s = self._type_alias    # used for on-the-fly template types
        else:
            s = self._type._name
        if self._type_args is not None:
            s = s + '<{}>'.format(self._type_args)
        if self._const:
            s = "const " + s
        if self._ptr:
            s += "*"
        if self._const_ptr:
            s += " const"
        if self._ptr_to_ptr:
            s += "*"
        if self._ref:
            s += "&"
        if self._funptr:
            s += ' (*{})' + self._argdecl
        else:
            s += ' {}'
        if self._array:
            if type(self._array_size) is str:
                s += '[{}]'.format(self._array_size)
            else:
                s += '[{}]'.format(str(self._array_size))
        return s




from EditDecorators import EditDecoratorsDialog
from PyDecorator import PyDecoratorDialog
from PyArgument import PyArgumentDialog
from PyClass import PyClassDialog
from PyFunction import PyFunctionDialog
from PyImport import PyImportDialog
from PyInheritance import PyInheritanceDialog
from PyMember import PyMemberDialog
from PyMemberMethod import PyMemberMethodDialog
from PyModule import PyModuleDialog
from PyPackage import PyPackageDialog
from PyVariable import PyVariableDialog



# -*- coding: utf-8 -*-
import mmap
import contextlib
import re

context_parser = None


class symbol(object):
    """This class represents a symbol for use in the
    parser generator. May be a terminal or not terminal
    symbol."""
    def __init__(self, name, parser=context_parser):
        """Initialize symbol"""
        self.name = name
        parser.AppendSymbol(self)
        super(symbol, self).__init__()


class terminal(symbol):
    """This class represents a terminal symbol"""
    def __init__(self, name, parser=context_parser):
        """Initialize a terminal symbol"""
        super(terminal, self).__init__(name, parser)


class literal(terminal):
    """This class represents a literal terminal simbol"""
    def __init__(self, charsequence, parser=context_parser):
        """Initialize a literal symbol"""
        name = 'LITERAL{}'.format(len(parser.literal) + 1)
        self.literal = charsequence
        super(literal, self).__init__(name, parser)


class nonterminal(symbol):
    """This class represents a non terminal symbol"""
    def __init__(self, name, parser=context_parser):
        """Initialize a non terminal symbol"""
        self._rules = []
        super(nonterminal, self).__init__(name, parser)

    def __ior__(self, rules, parser=context_parser):
        """Append rules"""
        rs = rules.split('|')
        # clean rules
        for r in rs:
            #analyze each rule
            rule = []
            ss = r.strip().split(' ')
            for s in ss:
                #analyze symbol: find it on parser
                assert(s in parser.symbol)
                rule.append(parser.symbol[s])
            self._rules.append(rule)

    def equivalent(self, other):
        """Check if the rules are the same for different symbols"""
        if self is other or type(other) != type(self):
            return False
        n = len(self._rules)
        if n != len(other._rules):
            return False
        return len([x for x in self._rules if x not in other._rules]) == 0

    def setRules(self, rules, parser=context_parser):
        """auxiliar method for assigning rules"""
        # The rules that we can handle can contain literal
        # symbols so we need to extract these symbols and
        # replace it by literal terminal symbols
        match = re.search(r"\'(((?<!\\)\\\'|[^\'])*)\'", rules)
        while match is not None:
            # get the literal text
            l = match.group(1)
            # find the literal in the parser
            if l not in parser.literal:
                #create a new literal
                lo = literal(l, parser)
                print("created new literal {l.name} = '{l.literal}'\n".format(l=lo))
            # ok, now, replace the literal
            rules = rules[:match.start(0)] + parser.literal[l].name + rules[match.end(0):]
            # and do next search
            match = re.search(r"\'(((?<!\\)\\\'|[^\'])*)\'", rules)
        # Ok, now, we need to parse each one of the rules but,
        # before doing so, we need to check the existence of subrules
        # embedded in the expressions. The possibilities are:
        # ...[subrule]... ...(subrule)...  and ...{subrule}...
        # These subrules are mapped to special symbols we called
        # option, subsymbol or sequence
        mapmatch = {
            'opt': (r"\[([^\]]*)\]", option),
            'sub': (r"\(([^\)]*)\)", subsymbol),
            'seq': (r"\{([^\}]*)\}", sequence)}
        for match_type in mapmatch:
            expr = mapmatch[match_type][0]
            hand = mapmatch[match_type][1]
            match = re.search(expr, rules)
            count = 0
            while match is not None:
                # get the rule text
                l = match.group(1)
                # ok, we create a subsymbol
                name = '{self.name}.{type}{count}'.format(self=self,
                    type=match_type, count=count)
                o = hand(name, parser)
                o.setRules(l, parser)
                count = count + 1
                # search in parser for equivalent symbol
                for sn in parser.symbol:
                    s = parser.symbol[sn]
                    if o.equivalent(s):
                        print("removed {o.name} by coincidence with {s.name}\n".format(
                            o=o, s=s))
                        del o
                        o = s
                        count = count - 1
                        break
                # replace the symbol
                rules = rules[:match.start(0)] + o.name + rules[match.end(0):]
                # do next search
                match = re.search(expr, rules)
                # Ok, now we find for subrules
        #Ok, now, we have pure rules and we separate it and process
        rs = rules.split('|')
        # clean rules
        self._rules = []
        for r in rs:
            #analyze each rule
            rule = []
            ss = r.strip().split(' ')
            for s in ss:
                if s not in parser.symbol:
                    print('Missing symbol {s}. Assumming terminal\n'.format(s=s))
                    terminal(s, parser)
                rule.append(parser.symbol[s])
            self._rules.append(rule)


class option(nonterminal):
    """Special nonterminal symbols"""
    def __init__(self, name, parser=context_parser):
        """Initialize """
        super(option, self).__init__(name, parser)


class subsymbol(nonterminal):
    """Special nonterminal symbols"""
    def __init__(self, name, parser=context_parser):
        """Initialize """
        super(subsymbol, self).__init__(name, parser)


class sequence(nonterminal):
    """Special nonterminal symbols"""
    def __init__(self, name, parser=context_parser):
        """Initialize """
        super(sequence, self).__init__(name, parser)


class parser(object):
    """This class defines a general parser generator.
    The generator analizes a sequence of terminal symbols
    for constructing the parser.

    Symbol kinds:

        * Terminal : output from scanner.
        * NonTerminal : expressed by semantic

    declaring rules:

        nonterminals = [
            nonterminal(parser,'const_expr'),
            nonterminal(parser,'mutable_expr'), ...]


        parser.const_expr = <expr1>|<expr2>|...|<exprN>

        each expression  is

            expression = "symbol"| "symbol expression"




    """
    def __init__(self, **kwargs):
        """Initialize parser"""
        global context_parser
        self.symbol = {}
        self.literal = {}
        context_parser = self
        super(parser, self).__init__()
        f = kwargs.get('file', None)
        if f:
            self.ReadFile(f)

    def AppendSymbol(self, symbol):
        """Appends the symbol"""
        self.symbol[symbol.name] = symbol
        if isinstance(symbol, terminal):
            setattr(self, symbol.name, property(
                lambda s: symbol,
                lambda s, v: eval("raise ValueError('terminal symbols are readonly')")))
            if type(symbol) is literal:
                self.literal[symbol.literal] = symbol
        elif isinstance(symbol, nonterminal):
            setattr(self, symbol.name, property(
                lambda s: symbol,
                lambda s, v: symbol.setRules(v, self)))

    def ReadFile(self, fname):
        """Read the parser definition from the file"""
        with open(fname, 'r') as f:
            with contextlib.closing(mmap.mmap(f.fileno(),
                0, access=mmap.ACCESS_READ)) as m:
                self.ReadString(m[:])

    def ReadString(self, sdef):
        """Read the parser definition from the string.
        The parser definition is written in the format:
            .<non-terminal> = rule 1|rule 2|

        The point character at beginning of each line
        is determinning the start of new non-terminal.
        """
        # split the string in non-terminal expressions
        expr = re.split(r'^\.', sdef, flags=re.M)

        #ok, create a map non-terminal -> definition
        nta = {}
        for k in expr:
            s = k.strip()
            if not len(s):
                continue
            i = s.find(':')
            if i < 1:
                print('invalid expression {}\n'.format(s))
                continue
            nts = s[:i].strip()
            rts = s[i + 1:].strip()
            if not re.match('^[a-zA-Z_][a-zA-Z0-9_]*$', nts):
                print('invalid non terminal identifier {}\n'.format(nts))
                continue
            if nts in nta:
                print('duplicated symbol {}\n'.format(nts))
            nta[nts] = rts
        #ok, first create the symbols
        vsy = {}
        for nts in nta:
            vsy[nts] = nonterminal(nts, parser=self)
            print('registered non terminal symbol {}'.format(nts))
            #now, we need to analize each entry: first, extract
            #literals
        #now, assign rules
        for nts in nta:
            vsy[nts].setRules(nta[nts], self)
        return






import os
from setuptools.command.install import install
from setuptools import setup, find_packages

class extra_install(install):
	def run(self):
		install.run(self)
		import sys
		#check about wxwidgets
		try:
			import wxversion
			wxversion.select('3.0')
		except:
			with open('install-wxpython.sh', 'w') as shell_script:
				shell_script.write("""#!/bin/sh
				echo 'Prerrequisites'
				apt-get --yes --force-yes install libgconf2-dev libgtk2.0-dev libgtk-3-dev mesa-common-dev libgl1-mesa-dev libglu1-mesa-dev libgstreamer0.10-dev libgstreamer-plugins-base0.10-dev libgconfmm-2.6-dev libwebkitgtk-dev python-gtk2
				mkdir -p external
				cd external
				echo "Downloading wxPython 3.0.2.0"
				wget http://downloads.sourceforge.net/project/wxpython/wxPython/3.0.2.0/wxPython-src-3.0.2.0.tar.bz2
				echo "Done"
				echo 'Uncompressing ...'
				tar -xjvf wxPython-src-3.0.2.0.tar.bz2
				echo 'Patching  ...'
				cd wxPython-src-3.0.2.0
				cd wxPython
				sed -i -e 's/PyErr_Format(PyExc_RuntimeError, mesg)/PyErr_Format(PyExc_RuntimeError, "%s\", mesg)/g' src/gtk/*.cpp contrib/gizmos/gtk/*.cpp
				python ./build-wxpython.py --build_dir=../bld --install
				ldconfig
				cd ../../..
				rm -rf external
			""")
			os.system("sh ./install-wxpython.sh")

def ftext(filename):
	"""Quick utility for reading a text file"""
	return open(os.path.join(os.path.dirname(__file__), filename)).read()

setup(
	name = 'beatle',
	version = '0.0.15',
	author = 'Mel Viso',
	author_email = 'melviso@telefonica.net',
	url = 'https://github.com/melviso/beatle',
	description = ('A development environment made in python'),
	license = 'GNU Public License',
	keywords = 'development python c++',
	long_description=ftext('README'),
	classifiers = [
		"Intended Audience :: Developers",
		"Development Status :: 2 - Pre-Alpha",
		"Topic :: Utilities",
		"License :: OSI Approved :: GNU General Public License (GPL)",
		"Natural Language :: English",
		"Operating System :: POSIX :: Linux",
		"Programming Language :: Python :: 2.7",
		"Topic :: Software Development :: Code Generators",
	],
	package_data = {'':['beatle/app/res/*.xpm','beatle/app/res/*.jpg','beatle/plugin/models/relation/standard/*.xml'],},
	include_package_data=True,
	packages=[
		'beatle',
		'beatle.activity',
		'beatle.activity.files',
		'beatle.activity.files.actions',
		'beatle.activity.files.ui',
		'beatle.activity.files.ui.dlg',
		'beatle.activity.files.ui.view',
		'beatle.activity.files.ui.pane',
		'beatle.activity.git',
		'beatle.activity.git.ui',
		'beatle.activity.git.ui.dlg',
		'beatle.activity.git.ui.view',
		'beatle.activity.models',
		'beatle.activity.models.ui',
		'beatle.activity.models.ui.dlg',
		'beatle.activity.models.ui.dlg.cc',
		'beatle.activity.models.ui.dlg.py',
		'beatle.activity.models.ui.pane',
		'beatle.activity.models.ui.view',
		'beatle.activity.targets',
		'beatle.activity.targets.ui',
		'beatle.activity.targets.ui.view',
		'beatle.activity.tasks',
		'beatle.activity.tasks.ui',
		'beatle.activity.tasks.ui.dlg',
		'beatle.activity.tasks.ui.view',
		'beatle.analytic',
		'beatle.analytic.sema',
		'beatle.app',
		'beatle.app.resources',
		'beatle.app.res',
		'beatle.app.res.16x16',
		'beatle.app.ui',
		'beatle.app.ui.ctrl',
		'beatle.app.ui.dlg',
		'beatle.app.ui.pane',
		'beatle.app.ui.tools',
		'beatle.app.ui.view',
		'beatle.builtin',
		'beatle.builtin.libraries',
		'beatle.builtin.libraries.stl',
		'beatle.builtin.libraries.stl.containers',
		'beatle.builtin.libraries.stl.utility',
		'beatle.ctx',
		'beatle.deco',
		'beatle.model',
		'beatle.model.cc',
		'beatle.model.decorator',
		'beatle.model.file',
		'beatle.model.git',
		'beatle.model.py',
		'beatle.model.tasks',
		'beatle.model.writer',
		'beatle.ostools',
		'beatle.pane',
		'beatle.plugin',
		'beatle.plugin.models',
		'beatle.plugin.models.relation',
		'beatle.plugin.models.relation.standard',
		'beatle.plugin.tools',
		'beatle.plugin.tools.ast_explorer',
		'beatle.plugin.tools.ast_explorer.res',
		'beatle.plugin.tools.log_explorer',
		'beatle.plugin.tools.web_browser',
		'beatle.plugin.tools.check_updates',
		'beatle.tran',
		'beatle.wxx',
		],
	cmdclass={'install': extra_install},
	install_requires = ['PyPDF2', 'GitPython', 'trepan', ],
	scripts=['script/beatle'],
)
	
	



# -*- coding: utf-8 -*-

# -*- coding: utf-8 -*-

import os
import signal
import wx
import wxx
import threading
from ctx import localcontext as context


class build(object):
    """Handler of debug session"""

    def __init__(self, project):
        """Init"""
        super(build, self).__init__()
        self._process = None
        self.make(project)

    def activateUI(self):
        """Activate the panes"""
        context.frame.CleanOutputPane()
        self.activateOutputPane()

    def deactivateUI(self):
        """Deactivate the ui"""
        frame = context.frame
        frame.m_mgr.Update()

    def activateOutputPane(self):
        """Activate the output pane"""
        frame = context.frame
        if frame.m_aux_panes == frame.m_aux_output_pane:
            return True
        try:
            i = max(i for i in range(0, frame.m_aux_panes.GetPageCount())
            if frame.m_aux_panes.GetPage(i) == frame.m_aux_output_pane)
            frame.m_aux_panes.SetSelection(i)
            return True
        except:
            pass
        return False

    def receiver_proc(self):
        """funcion que realiza el trabajo en el thread"""
        handler = context.frame.GetEventHandler()
        handler.AddPendingEvent(wxx.LoggerEvent('build ...'))
        line = ''
        for line in iter(self._process.stdout.readline, ""):
            handler.AddPendingEvent(wxx.LoggerEvent(line.strip()))
            self._process.stdout.flush()
            wx.YieldIfNeeded()
        handler.AddPendingEvent(wxx.LoggerEvent('end'))
        self._process.stdout.close()
        self._process.wait()
        self._process = None

    def force_stop(self):
        """Brute stop"""
        self.done = True
        if self._process:
            os.killpg(os.getpgid(self._process.pid), signal.SIGTERM)

    def make(self, project):
        """Start a debug operation"""
        if self._process:
            return False  # is already running
        path = os.getcwd()
        os.chdir(project.dir)
        server_command = 'make all'
        import subprocess
        try:
            self._process = subprocess.Popen(server_command,
                bufsize=1, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True,
                preexec_fn=os.setsid)
            threading.Thread(target=self.receiver_proc,).start()
        except:
            handler = context.frame.GetEventHandler()
            handler.AddPendingEvent(wxx.LoggerEvent('execution failed'))
            os.chdir(path)
            return False
        self.activateUI()
        wx.YieldIfNeeded()
        os.chdir(path)
        return True








# -*- coding: utf-8 -*-

from utils import import_once
from utils import isclass
from utils import cached_type
from utils import join_access
from utils import access_dict
from utils import store_access
from clipboard import clipboard

# import event identifiers
from proCxxGUI import ID_NEW_WORKSPACE
from proCxxGUI import ID_NEW_PROJECT
from proCxxGUI import ID_OPEN_PROJECT
from proCxxGUI import ID_IMPORT_PROJECT
from proCxxGUI import ID_SAVE_PROJECT
from proCxxGUI import ID_SAVE_WORKSPACE
from proCxxGUI import ID_CUT
from proCxxGUI import ID_EDIT_CONTEXT
from proCxxGUI import ID_EDIT_PROPERTIES
from proCxxGUI import ID_COPY
from proCxxGUI import ID_PASTE
from proCxxGUI import ID_DELETE
from proCxxGUI import ID_UNDO
from proCxxGUI import ID_REDO
from proCxxGUI import ID_COPY
from proCxxGUI import ID_EDIT_USER_SECTIONS
from proCxxGUI import ID_EDIT_OPEN
from proCxxGUI import ID_EDIT_CONTEXT
from proCxxGUI import ID_CLOSE_WORKSPACE
from proCxxGUI import ID_CLOSE_PROJECT


from proCxxGUI import NavigatorPane
from proCxxGUI import NewFolder
from proCxxGUI import NewProject
from proCxxGUI import ImportProject
from proCxxGUI import Import
from proCxxGUI import NewWorkspace
from proCxxGUI import TasksPane
from proCxxGUI import Preferences
from proCxxGUI import BuildTools
from proCxxGUI import BuildBinaries
from proCxxGUI import NewNote
from proCxxGUI import NewFile
from proCxxGUI import codeNavigator
from proCxxGUI import FullScreen
from proCxxGUI import Wait
from proCxxGUI import Working
from proCxxGUI import FindText
from proCxxGUI import FindInFiles
from proCxxGUI import PdfViewPane
from mainWindow import mainWindow

#from bmpTools import dumpBitmap
from bmpTools import PrefixBitmapList
from bmpTools import SuffixBitmapList
from bmpTools import GetBitmapMaxSize
from main import proCxx
from resources import GetBitmap
from splash import AppSplashScreen

from ui import edit
from ui import append_menuitem_copy
from ui import clone_mnu
from ui import set_menu_handlers
from ui import unset_menu_handlers



"""Subclass of NewClassDiagram, which is generated by wxFormBuilder."""

import wx
import wxx
import app.resources as rc
from activity.models.ui import ui as ui



# Implementing NewClassDiagram
class ClassDiagramDialog(ui.NewClassDiagram):
    """
    This dialog allows to create a new class diagram where to
    represent hierarchy and relationship between your project
    elements.
    """
    @wxx.SetInfo(__doc__)
    def __init__(self, parent, container):
        """Initializations"""
        super(ClassDiagramDialog, self).__init__(parent)
        self._container = container
        icon = wx.EmptyIcon()
        icon.CopyFromBitmap(rc.GetBitmap("classdiagram"))
        self.SetIcon(icon)
        self.m_textCtrl2.SetFocus()

    def Validate(self):
        """Validation"""
        self._name = self.m_textCtrl2.GetValue()
        if len(self._name) == 0:
            wx.MessageBox("Diagram name must not be empty", "Error",
                wx.OK | wx.CENTER | wx.ICON_ERROR, self)
            return False
        self._note = self.m_richText3.GetValue()
        return True

    def CopyAttributes(self, diag):
        """Transfer attributes to diagram"""
        diag._name = self._name
        diag.note = self._note

    def SetAttributes(self, diag):
        """Transfer attributes from existing diagram"""
        self._container = diag.parent
        self._name = diag.name
        self._note = diag.note
        self.m_textCtrl2.SetValue(self._name)
        self.m_richText3.SetValue(self._note)
        self.SetTitle("Edit diagram")

    def get_kwargs(self):
        """Returns arguments suitable for creating object"""
        return {'parent': self._container,
            'name': self._name, 'note': self._note}

    def OnCancel(self, event):
        """cancel event handler"""
        self.EndModal(wx.ID_CANCEL)

    def OnOK(self, event):
        """ok event handler"""
        if self.Validate():
            self.EndModal(wx.ID_OK)





# -*- coding: utf-8 -*-

"""
This class allows to set some common handlers and methods for
easy view declaration"""

import wx
import wxx
import tran
import app
from app.ui import dlg
from activity.models.ui import dlg as mdlg
import model
from ctx import localcontext as context


class BaseView(object):
    """Declares a base view class"""
    # class variables defines installed handlers
    workspace_handler = True
    project_handler = True
    clipboard_handler = True
    build_handler = False
    run_handler = False
    debug_handler = False

    # command ids: tree
    _leftKeyId = wx.Window.NewControlId()

    #command ids: build
    _buildProjectId = wx.Window.NewControlId()

    #command ids: run
    _runProjectId = wx.Window.NewControlId()
    _runFileId = wx.Window.NewControlId()
    _stopRunId = wx.Window.NewControlId()
    _addPathToConsoleId = wx.Window.NewControlId()

    #command ids: debug
    _toggleBreakpointId = wx.Window.NewControlId()
    _debugProjectId = wx.Window.NewControlId()
    _debugFileId = wx.Window.NewControlId()
    _debugContinueId = wx.Window.NewControlId()
    _debugStepOverId = wx.Window.NewControlId()
    _debugStepIntoId = wx.Window.NewControlId()
    _debugStepOutId = wx.Window.NewControlId()
    _debugStopId = wx.Window.NewControlId()

    #update_context : these properties will be used for update menus
    @property
    def can_edit_properties(self):
        """Are the properties of the selected element editable?"""
        if self.selected and not self.selected.read_only:
            return True
        return False

    def __init__(self, *args, **kwargs):
        """Initialize base view"""
        self._evtHandler = wx.PyEvtHandler()
        self._handler_installed = False
        self._registered_menus = {}
        self._registered_toolbars = []
        self._perspective_base = None
        self.selected = None
        super(BaseView, self).__init__(*args, **kwargs)

    def _create_menus(self):
        """Create the base view menus"""
        if self.project_handler:
            self._menuProject = wx.Menu()
            if self.build_handler:
                self._submenuBuild = wx.Menu()
                self._buildProject = wx.MenuItem(self._submenuBuild, self._buildProjectId, u"Project",
                    u"build project", wx.ITEM_NORMAL)
                self._buildProject.SetBitmap(wx.Bitmap(u"app/res/build.xpm", wx.BITMAP_TYPE_ANY))
                self._submenuBuild.AppendItem(self._buildProject)
                self._menuProject.AppendSubMenu(self._submenuBuild, u"Build")
            if self.run_handler:
                self._submenuRun = wx.Menu()
                # run project
                self._runProject = wx.MenuItem(self._submenuRun, self._runProjectId, u"Project",
                    u"run project", wx.ITEM_NORMAL)
                self._runProject.SetBitmap(wx.Bitmap(u"app/res/run.xpm", wx.BITMAP_TYPE_ANY))
                self._submenuRun.AppendItem(self._runProject)
                # run file
                self._runFile = wx.MenuItem(self._submenuRun, self._runFileId, u"File",
                    u"run current file", wx.ITEM_NORMAL)
                self._runFile.SetBitmap(wx.Bitmap(u"app/res/run_file.xpm", wx.BITMAP_TYPE_ANY))
                self._submenuRun.AppendItem(self._runFile)
                # stop run
                self._stopRun = wx.MenuItem(self._submenuRun, self._stopRunId, u"Stop",
                    u"stop project", wx.ITEM_NORMAL)
                self._stopRun.SetBitmap(wx.Bitmap(u"app/res/stop.xpm", wx.BITMAP_TYPE_ANY))
                self._submenuRun.AppendItem(self._stopRun)

                self._menuProject.AppendSubMenu(self._submenuRun, u"Run")

            if self.debug_handler:
                self._submenuDebug = wx.Menu()
                # breakpoint
                self._toggleBreakpoint = wx.MenuItem(self._submenuDebug, self._toggleBreakpointId, u"Breakpoint\tF9",
                    u"toggle breakpoint", wx.ITEM_CHECK)
                #self._toggleBreakpoint.SetBitmap(wx.Bitmap(u"res/breakpoint.xpm", wx.BITMAP_TYPE_ANY))
                self._submenuDebug.AppendItem(self._toggleBreakpoint)

                self._submenuDebug.AppendSeparator()
                # debug project
                self._debugProject = wx.MenuItem(self._submenuDebug, self._debugProjectId, u"Project",
                    u"debug project", wx.ITEM_NORMAL)
                self._debugProject.SetBitmap(wx.Bitmap(u"app/res/debug.xpm", wx.BITMAP_TYPE_ANY))
                self._submenuDebug.AppendItem(self._debugProject)
                # debug file
                self._debugFile = wx.MenuItem(self._submenuDebug, self._debugFileId, u"File",
                    u"debug file", wx.ITEM_NORMAL)
                self._debugFile.SetBitmap(wx.Bitmap(u"app/res/debug_file.xpm", wx.BITMAP_TYPE_ANY))
                self._submenuDebug.AppendItem(self._debugFile)
                # debug continue
                self._debugContinue = wx.MenuItem(self._submenuDebug, self._debugContinueId,
                    u"Continue\tF5", u"continue", wx.ITEM_NORMAL)
                self._debugContinue.SetBitmap(wx.Bitmap(u"app/res/continue.xpm", wx.BITMAP_TYPE_ANY))
                self._submenuDebug.AppendItem(self._debugContinue)
                # debug step over
                self._debugStepOver = wx.MenuItem(self._submenuDebug, self._debugStepOverId,
                    u"Step over\tF6", u"step over", wx.ITEM_NORMAL)
                self._debugStepOver.SetBitmap(wx.Bitmap(u"app/res/step_over.xpm", wx.BITMAP_TYPE_ANY))
                self._submenuDebug.AppendItem(self._debugStepOver)
                # debug step into
                self._debugStepInto = wx.MenuItem(self._submenuDebug, self._debugStepIntoId,
                    u"Step into\tF7", u"step into", wx.ITEM_NORMAL)
                self._debugStepInto.SetBitmap(wx.Bitmap(u"app/res/step_into.xpm", wx.BITMAP_TYPE_ANY))
                self._submenuDebug.AppendItem(self._debugStepInto)
                # debug step out
                self._debugStepOut = wx.MenuItem(self._submenuDebug, self._debugStepOutId,
                    u"Step out\tF8", u"step out", wx.ITEM_NORMAL)
                self._debugStepOut.SetBitmap(wx.Bitmap(u"app/res/step_out.xpm", wx.BITMAP_TYPE_ANY))
                self._submenuDebug.AppendItem(self._debugStepOut)
                # debug stop
                self._debugStop = wx.MenuItem(self._submenuDebug, self._debugStopId, u"Stop",
                    u"stop debug", wx.ITEM_NORMAL)
                self._debugStop.SetBitmap(wx.Bitmap(u"app/res/stop.xpm", wx.BITMAP_TYPE_ANY))
                self._submenuDebug.AppendItem(self._debugStop)

                self._menuProject.AppendSubMenu(self._submenuDebug, u"Debug")

            self._addPathToConsole = wx.MenuItem(self._menuProject, self._addPathToConsoleId, u"Add path to console",
                    u"add the project path to console", wx.ITEM_NORMAL)
            self._menuProject.AppendItem(self._addPathToConsole)

            self.RegisterMenu('Project', self._menuProject)

    def RegisterMenu(self, name, menu):
        """Appends a dynamic menu to the view, This menu will be
        shown only when the view becomes active"""
        self._registered_menus[name] = menu

    def RegisterToolbar(self, toolbar):
        """Appends a dynamic toolbar to the view, This toolbar will be
        shown only when the view becomes active"""
        self._registered_toolbars.append(toolbar)

    def BindSpecial(self, evtID, handler, source=None, id=-1, id2=-1):
        """Do a special binding for pluggable event handler and also internal event handler"""
        self._evtHandler.Bind(evtID, handler, source=source, id=id, id2=id2)
        self.Bind(evtID, handler, source=source, id=id, id2=id2)

    def _delete_toolbars(self):
        """Destroy associated toolbard"""
        while len(self._registered_toolbars):
            self._registered_toolbars.pop().Destroy()

    def _create_toolbars(self):
        """Create the associated toolbars"""
        # common elements
        if self.run_handler:
            self._barRun = wx.aui.AuiToolBar(self.frame, wx.ID_ANY, wx.DefaultPosition,
                wx.DefaultSize, wx.aui.AUI_TB_GRIPPER | wx.aui.AUI_TB_HORZ_LAYOUT)

            els = [self._runFile, self._runProject, self._stopRun]
            bar = self._barRun
            bar.SetMinSize(wx.Size(24, 24))
            for el in els:
                bar.AddTool(el.GetId(), el.GetLabel(), el.GetBitmap())
            bar.Realize()
            name = "run_toolbar_{}".format(self.__class__)
            self.frame.m_mgr.AddPane(bar, wx.aui.AuiPaneInfo().Name(name).Top().Hide().
                Caption(u"run").PinButton(True).Gripper().Dock().Resizable().
                DockFixed(False).Row(1).Position(0).Layer(10).ToolbarPane())
            self.RegisterToolbar(bar)

        if self.debug_handler:
            self._barDebug = wx.aui.AuiToolBar(self.frame, wx.ID_ANY, wx.DefaultPosition,
                wx.DefaultSize, wx.aui.AUI_TB_GRIPPER | wx.aui.AUI_TB_HORZ_LAYOUT)

            els = [self._debugFile, self._debugProject, self._debugContinue,
                self._debugStepInto, self._debugStepOver, self._debugStepOut,
                self._debugStop]
            bar = self._barDebug
            bar.SetMinSize(wx.Size(24, 24))
            for el in els:
                bar.AddTool(el.GetId(), el.GetLabel(), el.GetBitmap())
            bar.AddSeparator()
            bar.AddTool(self._toggleBreakpointId,
                "breakpoint",
                wx.Bitmap(u"app/res/breakpoint.xpm", wx.BITMAP_TYPE_ANY),
                "toggle breakpoint", wx.ITEM_CHECK
                )
            bar.Realize()
            name = "debug_toolbar_{}".format(self.__class__)
            self.frame.m_mgr.AddPane(bar, wx.aui.AuiPaneInfo().Name(name).Top().Hide().
                Caption(u"debug").PinButton(True).Gripper().Dock().Resizable().
                DockFixed(False).Row(1).Position(1).Layer(10).ToolbarPane())
            self.RegisterToolbar(bar)

    def _bind_events(self):
        """Connect common handlers"""
        if self.workspace_handler:
            self.BindSpecial(wx.EVT_MENU, self.OnSaveWorkspace, id=app.ID_SAVE_WORKSPACE)
            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateSaveWorkspace, id=app.ID_SAVE_WORKSPACE)
            self.BindSpecial(wx.EVT_MENU, self.OnCloseWorkspace, id=app.ID_CLOSE_WORKSPACE)
            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateCloseWorkspace, id=app.ID_CLOSE_WORKSPACE)
            self.BindSpecial(wx.EVT_MENU, self.OnCloseProject, id=app.ID_CLOSE_PROJECT)
            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateCloseProject, id=app.ID_CLOSE_PROJECT)

        if self.build_handler:
            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateBuildProject, id=self._buildProjectId)
            self.BindSpecial(wx.EVT_MENU, self.OnBuildProject, id=self._buildProjectId)

        if self.project_handler:
            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateSaveProject, id=app.ID_SAVE_PROJECT)
            self.BindSpecial(wx.EVT_MENU, self.OnSaveProject, id=app.ID_SAVE_PROJECT)

            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateAddPathToConsole, id=self._addPathToConsoleId)
            self.BindSpecial(wx.EVT_MENU, self.OnAddPathToConsole, id=self._addPathToConsoleId)

        if self.run_handler:
            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateRunProject, id=self._runProjectId)
            self.BindSpecial(wx.EVT_MENU, self.OnRunProject, id=self._runProjectId)

            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateRunFile, id=self._runFileId)
            self.BindSpecial(wx.EVT_MENU, self.OnRunFile, id=self._runFileId)

            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateStopRun, id=self._stopRunId)
            self.BindSpecial(wx.EVT_MENU, self.OnStopRun, id=self._stopRunId)

        if self.debug_handler:
            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateToggleBreakpoint, id=self._toggleBreakpointId)
            self.BindSpecial(wx.EVT_MENU, self.OnToggleBreakpoint, id=self._toggleBreakpointId)

            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateDebugProject, id=self._debugProjectId)
            self.BindSpecial(wx.EVT_MENU, self.OnDebugProject, id=self._debugProjectId)

            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateDebugFile, id=self._debugFileId)
            self.BindSpecial(wx.EVT_MENU, self.OnDebugFile, id=self._debugFileId)

            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateDebugContinue, id=self._debugContinueId)
            self.BindSpecial(wx.EVT_MENU, self.OnDebugContinue, id=self._debugContinueId)

            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateDebugStepOver, id=self._debugStepOverId)
            self.BindSpecial(wx.EVT_MENU, self.OnDebugStepOver, id=self._debugStepOverId)

            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateDebugStepInto, id=self._debugStepIntoId)
            self.BindSpecial(wx.EVT_MENU, self.OnDebugStepInto, id=self._debugStepIntoId)

            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateDebugStepOut, id=self._debugStepOutId)
            self.BindSpecial(wx.EVT_MENU, self.OnDebugStepOut, id=self._debugStepOutId)

            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateDebugStop, id=self._debugStopId)
            self.BindSpecial(wx.EVT_MENU, self.OnDebugStop, id=self._debugStopId)

        self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateEditProperties, id=app.ID_EDIT_PROPERTIES)
        self.BindSpecial(wx.EVT_MENU, self.OnEditProperties, id=app.ID_EDIT_PROPERTIES)

        self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateDelete, id=app.ID_DELETE)
        self.BindSpecial(wx.EVT_MENU, self.OnDelete, id=app.ID_DELETE)

        self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateNewProject, id=app.ID_NEW_PROJECT)
        self.BindSpecial(wx.EVT_MENU, self.OnNewProject, id=app.ID_NEW_PROJECT)

        self.BindSpecial(wx.EVT_MENU, self.OnOpenProject, id=app.ID_OPEN_PROJECT)
        self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateOpenProject, id=app.ID_OPEN_PROJECT)

        self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateImportProject, id=app.ID_IMPORT_PROJECT)
        self.BindSpecial(wx.EVT_MENU, self.OnImportProject, id=app.ID_IMPORT_PROJECT)

        self.BindSpecial(wx.EVT_MENU, self.OnShowViewToolbars, id=context.frame.m_view_showToolbars.GetId())

        # if the view has transactional behavior, activate edit handlers
        if self.clipboard_handler:
            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateEditCopy, id=app.ID_COPY)
            self.BindSpecial(wx.EVT_MENU, self.OnEditCopy, id=app.ID_COPY)

            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateEditCut, id=app.ID_CUT)
            self.BindSpecial(wx.EVT_MENU, self.OnEditCut, id=app.ID_CUT)

            self.BindSpecial(wx.EVT_UPDATE_UI, self.OnUpdateEditPaste, id=app.ID_PASTE)
            self.BindSpecial(wx.EVT_MENU, self.OnEditPaste, id=app.ID_PASTE)

        self.BindSpecial(wx.EVT_MENU, self.frame.OnMRUFile, None, wx.ID_FILE1, wx.ID_FILE9)

        # bind special navigation facility
        self.Bind(wx.EVT_MENU, self.TreeLeftKey, id=BaseView._leftKeyId)

    #default handlers
    def OnUpdateRunProject(self, event):
        """default event handler"""
        event.Enable(False)

    def OnUpdateRunFile(self, event):
        """default event handler"""
        event.Enable(False)

    def OnUpdateStopRun(self, event):
        """default event handler"""
        event.Enable(False)

    def OnUpdateToggleBreakpoint(self, event):
        """default event handler"""
        event.Enable(False)

    def OnUpdateDebugProject(self, event):
        """default event handler"""
        event.Enable(False)

    def OnUpdateDebugFile(self, event):
        """default event handler"""
        event.Enable(False)

    def OnUpdateDebugContinue(self, event):
        """default event handler"""
        event.Enable(False)

    def OnUpdateDebugStepOver(self, event):
        """default event handler"""
        event.Enable(False)

    def OnUpdateDebugStepInto(self, event):
        """default event handler"""
        event.Enable(False)

    def OnUpdateDebugStepOut(self, event):
        """default event handler"""
        event.Enable(False)

    def OnUpdateDebugStop(self, event):
        """default event handler"""
        event.Enable(False)

    def OnRunProject(self, event):
        """default event handler"""
        pass

    def OnRunFile(self, event):
        """default event handler"""
        pass

    def OnStopRun(self, event):
        """default event handler"""
        pass

    def OnDebugProject(self, event):
        """default event handler"""
        pass

    def OnToggleBreakpoint(self, event):
        """default event handler"""
        pass

    def OnDebugFile(self, event):
        """default event handler"""
        pass

    def OnDebugContinue(self, event):
        """default event handler"""
        pass

    def OnDebugStepOver(self, event):
        """default event handler"""
        pass

    def OnDebugStepInto(self, event):
        """default event handler"""
        pass

    def OnDebugStepOut(self, event):
        """default event handler"""
        pass

    def OnDebugStop(self, event):
        """default event handler"""
        pass

    def OnUpdateEditProperties(self, event):
        """Handle update edit properties"""
        event.Enable(self.can_edit_properties)

    def OnEditProperties(self, event):
        """Handle edit properties method"""
        app.ui.edit(context.frame, self.selected)

    def OnUpdateSaveWorkspace(self, event):
        """Handle update save project"""
        event.Enable(bool(self.selected and self.selected.workspace))

    def OnSaveWorkspace(self, event):
        """Handle save project"""
        wrk = self.selected and self.selected.workspace
        if wrk is None:
            return
        context.app.SaveWorkspace(wrk)

    def OnUpdateCloseWorkspace(self, event):
        """Handle update close workspace"""
        event.Enable(bool(self.selected and self.selected.workspace))

    @tran.TransactionalMethod('close workspace')
    def OnCloseWorkspace(self, event):
        """Handle save project"""
        wrk = self.selected and self.selected.workspace
        if wrk is None:
            return
        tran.TransactionStack.DoUnload(wrk)
        return True

    def OnUpdateCloseProject(self, event):
        """Handle update close workspace"""
        event.Enable(bool(self.selected and self.selected.project))

    @tran.TransactionalMethod('close project')
    def OnCloseProject(self, event):
        """Handle save project"""
        prj = self.selected and self.selected.project
        if prj is None:
            return
        tran.TransactionStack.DoUnload(prj)
        return True

    def OnUpdateBuildProject(self, event):
        """Update build project"""
        s = self.selected
        event.Enable(bool(s and s.project and s.project.buildable))

    def OnBuildProject(self, event):
        """Update build project"""
        pass

    def OnUpdateSaveProject(self, event):
        """Update save project method"""
        event.Enable(bool(self.selected and self.selected.project))

    def OnSaveProject(self, event):
        """Handle save project"""
        obj = self.selected
        project = obj.project
        if project is None:
            return
        context.app.SaveProject(project)

    def OnUpdateAddPathToConsole(self, event):
        """Update save project method"""
        event.Enable(False)

    def OnAddPathToConsole(self, event):
        """Update save project method"""
        pass

    def OnUpdateNewProject(self, event):
        """Update add project"""
        obj = self.selected
        event.Enable(bool(obj and obj.inner_project_container))

    @tran.TransactionalMethod('new project {}')
    @wxx.CreationDialog(mdlg.ProjectDialog, model.Project)
    def OnNewProject(self, event):
        """Handle new project command"""
        return (context.frame,
            self.selected and self.selected.inner_project_container)

    def OnUpdateOpenProject(self, event):
        """Handle update edit properties"""
        event.Enable(bool(self.selected and self.selected.inner_project_container))

    @tran.TransactionalMethod('open project')
    def OnOpenProject(self, event):
        """Handle open project method"""
        path = wx.EmptyString
        v = self.selected
        if v:
            workspace = v.workspace
            if workspace:
                path = workspace._dir
        else:
            workspace = None
        dialog = wx.FileDialog(self, "Select project file", path,
            wx.EmptyString, "Project files (*.pcc)|*.pcc", wx.FD_OPEN)
        if dialog.ShowModal() != wx.ID_OK:
            return None
        project = context.app.LoadProject(dialog.GetPath(), workspace)
        if project:
            tran.TransactionStack.DoLoad(project)
            if not workspace:
                mru = context.mru
                mru.AddFileToHistory(fname)
            project.SetModified(False)
            # check classes
            for cls in project.classes:
                if cls.errorCheckDerivatives():
                    project.SetModified(True)
        return project


    def OnUpdateImportProject(self, event):
        """Update import procjec"""
        obj = self.selected
        event.Enable(bool(obj and obj.workspace))

    @tran.TransactionalMethod('import project')
    def OnImportProject(self, event):
        """Handle import project event"""
        dialog = dlg.ImportProjectDialog(context.frame, self.selected.workspace)
        if dialog.ShowModal() == wx.ID_OK:
            # create modeless dialog
            progress = dlg.WaitDialog(context.frame)
            progress.InitDialog()
            progress.Show(True)
            progress.Centre()
            progress.Refresh()
            progress.Update()
            v = dialog.DoImport(progress)
            progress.Close()
            return v
        return False

    def install_handler(self):
        """Installs the event handler"""
        if self._handler_installed:
            return False  # already installed
        context.frame.PushEventHandler(self._evtHandler)
        self._handler_installed = True
        return True  # ok, installed

    def remove_handler(self):
        """Remove the event handler"""
        if not self._handler_installed:
            return False  # already removed
        # pop event handlers until found this
        other_event_handlers = []
        evth = context.frame.PopEventHandler()
        try:
            while evth != self._evtHandler:
                other_event_handlers.append(evth)
                evth = context.frame.PopEventHandler()
            if evth == self._evtHandler:
                self._handler_installed = False
            result = True
        except:
            print "Fatal error : missing event handler"
            self._handler_installed = False
            result = False
        # reinstall the event handlers chain
        while len(other_event_handlers):
            context.frame.PushEventHandler(other_event_handlers.pop())
        return result

    def NotifyShow(self):
        """Called where the git view is show"""
        if not self.install_handler():
            return  # avoid extra calls
        for menu in self._registered_menus:
            self.frame.AddMainMenu(menu, self._registered_menus[menu])
        if self._perspective_base is None:
            # if there is not a base perspective, we create one
            for toolbar in self._registered_toolbars:
                context.frame.m_mgr.GetPane(toolbar).Show()
                toolbar.Realize()
                context.frame.m_mgr.Update()
            self._perspective_base = context.frame.m_mgr.SavePerspective()
        cls = type(self)
        if cls.perspective:
            context.frame.m_mgr.LoadPerspective(cls.perspective)

    def NotifyHide(self):
        """Called where the git view is hidden"""
        if not self.remove_handler():
            return  # avoid extra calls
        # store current perspective for later use
        cls = type(self)
        cls.perspective = context.frame.m_mgr.SavePerspective()
        for menu in self._registered_menus:
            self.frame.RemoveMainMenu(menu, self._registered_menus[menu])
        for toolbar in self._registered_toolbars:
            context.frame.m_mgr.GetPane(toolbar).Hide()
            toolbar.Realize()
            context.frame.m_mgr.Update()

    def OnShowViewToolbars(self, event):
        """Show all the view toolbars"""
        context.frame.m_mgr.LoadPerspective(self._perspective_base)





"""Subclass of Progress, which is generated by wxFormBuilder."""

import wx
import wxx
from activity.git.ui import ui as ui
import app.resources as rc
from git import RemoteProgress

class commm(RemoteProgress):

    def __init__(self, dialog):
        self.dialog = dialog
        super(commm, self).__init__()

    def update(self, op_code, cur_count, max_count=None, message=''):
        self.dialog.update(self, op_code, cur_count, max_count, message)


# Implementing Progress
class ProgressDialog(ui.Progress):
    """
    This dialogs shows the progress of the
    current git operation.

    """
    @wxx.SetInfo(__doc__)
    def __init__(self, parent):
        """Initialize the dialog"""
        super(ProgressDialog, self).__init__(parent)
        self.m_sdbSizer12OK.Enable(False)

    def link(self):
        """create a logger"""
        self._link = commm(self)
        return self._link

    def update(self, op_code, cur_count, max_count=None, message=''):
        """Update info from operation"""
        s = '{op_code}:[{cur_count} ({percent})] {message}\r'.format(
            op_code=op_code,
            cur_count=cur_count,
            percent=cur_count / (max_count or 100.0),
            message=message or "NO MESSAGE")
        self.m_text.AppendText(s)
        wx.YieldIfNeeded()

    def Finished(self):
        """Enables closing the dialog"""
        self.m_sdbSizer12OK.Enable(True)
        del self._link

    def OnOK(self, event):
        """Closes the dialog"""
        self.Destroy()




# -*- coding: utf-8 -*-


class writer(object):
    """Implements a generic C++ writer."""
    def __init__(self, language="c++"):
        """initialize the writer class"""
        self.cident = ""
        self._language = language
        super(writer, self).__init__()

    @classmethod
    def for_file(cls, fhandle, language="c++"):
        """method for new writer"""
        from model.writer import filewriter
        return filewriter(fhandle, language)

    @classmethod
    def for_mem(cls):
        """method for new writer"""
        from model.writer import memwriter
        return memwriter()

    def ident(self):
        """Increase ident size"""
        self.cident += "\t"

    def unident(self):
        """Increase ident size"""
        self.cident = self.cident[:-1]

    def openbrace(self, text=''):
        """Opens a brace"""
        if self._language == 'c++':
            self.writeln("{}{{".format(text))
        else:
            self.writeln("{}".format(text))
        self.ident()

    def closebrace(self, text=''):
        """Closes a brace"""
        self.unident()
        if self._language == 'c++':
            self.writeln("}}{}".format(text))
        else:
            self.writeln("{}".format(text))

    def writecomment(self, text):
        """Write a multiline comment respecting nesting"""
        self.opencomment()
        if type(text) is not str:  # avoid null fields
            text = str(text)
        for x in text.split('\n'):
            self.writeln('{}'.format(x.strip()))
        self.closecomment()

    def writeblock(self, text):
        """Write a text increasing nesting as required"""
        for x in text.split('\n'):
            self.writeln(x)

    def opencomment(self):
        """Open a c comment"""
        if self._language == 'c++':
            self.writeln("/*")
        if self._language == 'python':
            self.writeln("\"\"\"")

    def closecomment(self):
        """Close a c commen"""
        if self._language == 'c++':
            self.writeln("*/")
        elif self._language == 'python':
            self.writeln("\"\"\"")



# -*- coding: utf-8 -*-

# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
"""
Created on Sun Dec 15 23:24:30 2013

@author: mel
"""

import os
import sys
import wx
import model
from ctx import theContext as context
from model.writer import writer
import app.resources as rc
import traceback
import tran


class Package(model.TComponent):
    """Implements a Module representation"""
    class_container = True
    folder_container = True
    diagram_container = True
    namespace_container = False
    function_container = True
    variable_container = True
    enum_container = False
    import_container = True
    package_container = True
    module_container = True

    def __init__(self, **kwargs):
        """Initialization"""
        self._lastSrcTime = None
        self._lastHdrTime = None
        self._entry = False
        self._file_obj = None  # vincle with file object
        super(Package, self).__init__(**kwargs)
        # La forma mas simple de hacer las cosas bien
        # no es pasar menos trabajo. Creamos el modulo __init__
        if not kwargs.get('noinit', False):
            model.py.Module(parent=self, name='__init__', readonly=True)
        self.ExportPythonCodeFiles()

    def set_file(self, file_obj):
        """Sets the vincle with file object"""
        self._file_obj = file_obj

    def Delete(self):
        """delete the object"""
        if hasattr(self, '_file_obj') and self._file_obj:
            self._file_obj.Delete()
        super(Package, self).Delete()

    def get_kwargs(self):
        """Returns the kwargs needed for this object"""
        return super(Package, self).get_kwargs()

    @property
    def key_file(self):
        """This method returns the key file. For modules, is
        the implementation file. For packages, is the __init__ file."""
        fname = os.path.join(self.dir, '__init__.py')
        return os.path.realpath(fname)

    @property
    def dir(self):
        """returns the absolute path"""
        d = os.path.join(self.parent.dir, self._name)
        return os.path.realpath(d)

    def OnUndoRedoRemoving(self):
        """Handle OnUndoRedoRemoving, prevent generating files"""
        stack = tran.TransactionStack
        method = self.ExportPythonCodeFiles.inner
        if method not in stack.delayedCallsFiltered:
            stack.delayedCallsFiltered.append(method)
        super(Package, self).OnUndoRedoRemoving()

    @property
    def abs_path(self):
        """return the absolute path of __init__"""
        return os.path.join(self.dir, '__init__.py'.format(self=self))

    @tran.DelayedMethod
    def ExportPythonCodeFiles(self, force=False, logger=None):
        """does source generation"""
        # Any package is just a directory. And an init file
        package_dir = self.dir
        logger and logger.AppendReportLine('exporting package {}'.format(self._name), wx.ICON_INFORMATION)
        if not os.path.exists(package_dir):
            os.makedirs(package_dir)
            if not os.path.exists(package_dir):
                e = "Failed creating package directory " + package_dir
                if logger is not None:
                    logger.AppendReportLine(e, wx.ICON_ERROR)
                else:
                    wx.MessageBox(e, "Error",
                        wx.OK | wx.CENTER | wx.ICON_ERROR, context.frame)
                return False
        #write the init file
        init_file = self.abs_path
        try:
            f = open(init_file, 'w')
            fw = writer.for_file(f, 'python')

            #first thing: generate module comment
            fw.writeln("# -*- coding: utf-8 -*-")
            fw.writecomment(self._note)

            fw.writenl()
            #declare the orderer collections of things
            #that would be exported
            from Import import Import
            from Data import Data
            from Function import Function
            from Class import Class

            # write explicit imports
            fw.writeln("#imports")
            for obj in self[Import]:
                obj.ExportPythonCode(fw)

            #now, for all the exported classes not inside nested package
            package_element = lambda x: x.inner_package == self
            for obj in self(Class, filter=package_element, cut=True):
                if not obj._export:
                    continue
                fw.writeln("from {obj.file_container} import {obj.context_name}".format(obj=obj))

            top = lambda x: not x.inner_module and not x.parent.inner_class
            fw.writenl(2)
            fw.writeln("#data")
            for obj in self(Data):
                obj.ExportPythonCode(fw)

            fw.writenl(2)
            fw.writeln("#functions")
            for obj in self(Function):
                obj.ExportPythonCode(fw)

            fw.writenl(2)
            fw.writeln("#classes")
            for obj in self(Class, filter=top, cut=True):
                obj.ExportPythonCode(fw)
            f.close()

        except Exception as inst:
            traceback.print_exc(file=sys.stdout)
            print type(inst)     # the exception instance
            print inst.args      # arguments stored in .args
            print inst
            if logger:
                logger.AppendReportLine('failed opening {}'.format(init_file), wx.NOT_FOUND)
            return False

        # do the same as for projects

        for obj in self.packages:
            if obj.parent.inner_package != self:
                continue
            obj.ExportPythonCodeFiles(force, logger)
        for obj in self.modules:
            if obj.parent.inner_package != self:
                continue
            obj.ExportPythonCodeFiles(force, logger)

    @property
    def bitmap_index(self):
        """Index of tree image"""
        return rc.GetBitmapIndex("py_package")

    def _local_import(self, name):
        """This method finds for a module or package (in this order)
        matching the name of the import, inside the nested directory of
        the module. Please take consideration of that
        the name must be simple, without dots. Returns the module or package
        or None if not found"""
        container = self
        #if the package holds __init__, the import is translated
        #to this file
        f = lambda x: x != self and x.inner_package == container
        m = dict([(x._name, x) for x in self(model.py.Module, filter=f, cut=True)])
        if '__init__' in m:
            return m['__init__']
        if name in m:
            return m[name]
        f = lambda x: x.parent.inner_package == container
        for x in self(model.py.Package, filter=f, cut=True):
            if x._name == name:
                return x
        return None

    @property
    def inner_class(self):
        """Get the innermost class container"""
        return None




from Argument import ArgumentDialog
from Class import ClassDialog
from Constructor import ConstructorDialog
from Destructor import DestructorDialog
from Enum import EnumDialog
from EnumItem import EnumItemDialog
from Function import FunctionDialog
from Inheritance import InheritanceDialog
from IsClassMethods import IsClassMethods
from Member import MemberDialog
from MemberMethod import MemberMethodDialog
from Module import ModuleDialog
from Namespace import NamespaceDialog
from Relation import RelationDialog
from SelectClasses import SelectClassesDialog
from SelectContexts import SelectContexts
from SelectFriends import SelectFriendsDialog
from SelectLibraries import SelectLibrariesDialog
from Type import TypeDialog
from UserSections import UserSectionsDialog
from Variable import VariableDialog
from VirtualMethods import VirtualMethodsDialog
from BoostPythonModule import BoostPythonModuleDialog
from Navigator import NavigatorDialog



# -*- coding: utf-8 -*-
"""
Created on Sun Dec 22 22:08:46 2013

@author: mel
"""
import wx
import model
import app.resources as rc
from ctx import theContext as context
import tran


class Function(model.TComponent):
    """Implements member function"""
    context_container = True
    argument_container = True

    # visual methods
    @tran.TransactionalMethod('move function {}')
    def drop(self, to):
        """Drops datamember inside project or another folder """
        target = to.inner_function_container
        if not target or self.project != target.project:
            return False  # avoid move classes between projects
        index = 0
        tran.TransactionalMoveObject(
            object=self, origin=self.parent, target=target, index=index)
        return True

    def __init__(self, **kwargs):
        """Initialization"""
        self._content = kwargs.get('content', "")
        super(Function, self).__init__(**kwargs)
        k = self.inner_module or self.inner_package
        if k:
            k.ExportPythonCodeFiles()

    def Delete(self):
        """Handle delete"""
        k = self.inner_module or self.inner_package
        super(Function, self).Delete()
        if k:
            k.ExportPythonCodeFiles()

    def get_kwargs(self):
        """Returns the kwargs needed for this object"""
        kwargs = {}
        kwargs['content'] = self._content
        kwargs.update(super(Function, self).get_kwargs())
        return kwargs

    def WriteCode(self, f):
        """Write code to file"""
        pass

    def OnUndoRedoChanged(self):
        """Update from app"""
        super(Function, self).OnUndoRedoChanged()
        if hasattr(self, '_pane') and not self._pane is None:
            book = context.frame.docBook
            index = book.GetPageIndex(self._pane)
            book.SetPageText(index, self.tab_label)
            #book.SetPageBitmap(index, self.GetTabBitmap())
            if self._pane.m_editor.GetText() != self._content:
                self._pane.m_editor.SetText(self._content)

    def OnUndoRedoRemoving(self, root=True):
        """Prepare for delete"""
        if hasattr(self, '_pane') and not self._pane is None:
            book = context.frame.docBook
            p = self._pane
            delattr(self, '_pane')
            p.Commit()
            p.PreDelete()    # avoid gtk-critical
            self._paneIndex = book.GetPageIndex(p)
            book.DeletePage(self._paneIndex)
        super(Function, self).OnUndoRedoRemoving(root)

    def OnUndoRedoAdd(self):
        """Restore object from undo"""
        from activity.models.ui import pane
        super(Function, self).OnUndoRedoAdd()
        if hasattr(self, '_paneIndex'):
            book = context.frame.docBook
            p = pane.PyMethodPane(book, context.frame, self)
            self._pane = p
            book.InsertPage(self._paneIndex, p, self.tab_label,
                False, self.bitmap_index)
            delattr(self, '_paneIndex')

    def ExportPythonCode(self, wf):
        """Write code"""
        for deco in self[model.py.Decorator]:
            deco.ExportPythonCode(wf)
        wf.openbrace(self.declare)
        wf.writecomment(self._note)
        wf.writeblock(self._content)
        wf.closebrace()
        wf.writenl()

    def GetTabBitmap(self):
        """Get the bitmap for tab control"""
        return rc.GetBitmap("py_function")

    @property
    def bitmap_index(self):
        """Index of tree image"""
        return rc.GetBitmapIndex("py_function")

    @property
    def tree_label(self):
        """Get tree label"""
        from Argument import Argument
        from ArgsArgument import ArgsArgument
        from KwArgsArgument import KwArgsArgument
        alist = ', '.join(arg.label for arg in
            self[Argument] + self[ArgsArgument] + self[KwArgsArgument])
        return '{self._name}({alist})'.format(self=self, alist=alist)

    @property
    def tab_label(self):
        """Get tab label"""
        from Argument import Argument
        from ArgsArgument import ArgsArgument
        from KwArgsArgument import KwArgsArgument
        alist = ', '.join(arg.label for arg in
            self[Argument] + self[ArgsArgument] + self[KwArgsArgument])
        return '{self._name}({alist})'.format(self=self, alist=alist)

    @property
    def declare(self):
        """Get tree label"""
        from Argument import Argument
        from ArgsArgument import ArgsArgument
        from KwArgsArgument import KwArgsArgument
        alist = ', '.join(arg.label for arg in
            self[Argument] + self[ArgsArgument] + self[KwArgsArgument])
        return 'def {self._name}({alist}):'.format(self=self, alist=alist)

    #def GetFullLabel(self):
    #    """Gets tab label"""
    #    return self.label

    def ExistArgumentNamed(self, name):
        """Returns information about argument existence"""
        from Argument import Argument
        from ArgsArgument import ArgsArgument
        from KwArgsArgument import KwArgsArgument
        return name in [x._name for x in self(Argument, ArgsArgument, KwArgsArgument)]




# -*- coding: utf-8 -*-
"""
Created on Sun Dec 15 23:24:30 2013

@author: mel
"""
import model
import model.decorator as ctx
import app.resources as rc
import tran


class Folder(model.TComponent):
    """Implements a Folder representation"""
    class_container = True
    context_container = True
    folder_container = True
    diagram_container = True
    module_container = True
    namespace_container = True
    function_container = True
    variable_container = True
    member_container = True

    # visual methods
    @tran.TransactionalMethod('move folder {}')
    def drop(self, to):
        """Drops class inside project or another folder """
        target = to.inner_folder_container
        if not target or self.inner_class != target.inner_class or self.project != target.project:
            return False  # avoid move classes between projects
        index = 0
        tran.TransactionalMoveObject(
            object=self, origin=self.parent, target=target, index=index)
        return True

    def __init__(self, **kwargs):
        """Initialization"""
        super(Folder, self).__init__(**kwargs)
        self.update_container()

    def update_container(self):
        """Update the container info"""
        self.class_container = self._parent.class_container
        self.context_container = self._parent.context_container
        self.folder_container = self._parent.folder_container
        self.diagram_container = self._parent.diagram_container
        self.module_container = self._parent.module_container
        self.namespace_container = self._parent.namespace_container
        self.function_container = self._parent.function_container
        self.variable_container = self._parent.variable_container
        self.member_container = self._parent.member_container
        self.import_container = self._parent.import_container

    def get_kwargs(self):
        """Returns the kwargs needed for this object"""
        return super(Folder, self).get_kwargs()

    @ctx.ContextDeclaration()
    def WriteDeclaration(self, pf):
        """Write the folder declaration"""
        self.WriteComment(pf)
        #we will follow and respect the visual order
        child_types = [model.Folder, model.cc.Namespace,
            model.cc.Class, model.cc.Enum, model.cc.MemberData,
            model.cc.Constructor, model.cc.MemberMethod,
            model.cc.Destructor, model.cc.Module,
            model.cc.Function, model.cc.Data]
        for type in child_types:
            if len(self[type]):
                for item in self[type]:
                    item.WriteDeclaration(pf)

    @property
    def bitmap_index(self):
        """Index of tree image"""
        return rc.GetBitmapIndex("folder")

    @property
    def bitmap_open_index(self):
        """Index of tree image"""
        return rc.GetBitmapIndex("folder_open")

    def ExistMemberDataNamed(self, name):
        """Check recursively about the existence of nested child member"""
        from MemberData import MemberData
        return self.hasChild(name=name, type=MemberData)

    def AddMemberData(self, name):
        """Add a member element"""
        from MemberData import MemberData
        member = MemberData(name, self)
        return member

    @property
    def nested_classes(self):
        """Returns the list of nested classes (including self)"""
        if type(self.parent) not in [model.Folder, model.cc.Class]:
            return []
        return self.parent.nested_classes



"""Transaction stack definition"""
import copy


def copy_temps(obj):
    """Copy with temps"""
    s = copy.copy(obj)
    state = dict(obj.__dict__)
    if '_pane' in state:
        s._pane = obj._pane
    if '_paneIndex' in state:
        s._paneIndex = obj._paneIndex
    if '_parentIndex' in state:
        s._parentIndex = obj._parentIndex
    if '_elements' in state:
        s._elements = copy.copy(obj._elements)
    if '_ancestors' in state:
        s._ancestors = copy.copy(obj._ancestors)
    if '_derivatives ' in state:
        s._derivatives = copy.copy(obj._derivatives)
    if '_fromRelation ' in state:
        s._fromRelation = copy.copy(obj._fromRelation)
    if '_toRelation ' in state:
        s._toRelation = copy.copy(obj._toRelation)
    if '_contexts' in state:
        s._contexts = copy.copy(obj._contexts)
    return s


def respect_visuals_from(oldobj, newobj):
    """Respect visual changes"""
    state = dict(oldobj.__dict__)
    if '_pane' in state:
        newobj._pane = oldobj._pane
    else:
        newobj._pane = None


class Transaction(object):
    """Implements a single transaction"""
    def __init__(self, name):
        """Initialice transaction"""
        self._key = None
        self._name = name
        self._operations = []

    def Dispose(self):
        """Notify all transaction componentes that it will be"""
        for op in self._operations:
            op.Dispose()

    @property
    def key(self):
        """return the key"""
        return self._key

    def SetKey(self, key):
        """sets the key"""
        self._key = key

    def Rollback(self):
        """Rollback transaction"""
        while len(self._operations) > 0:
            self._operations.pop().Rollback()

    def Undo(self):
        """Undoes transaction"""
        redo_operations = []
        while len(self._operations) > 0:
            redo_operations.append(self._operations.pop().Undo())
        self._operations = redo_operations
        return self

    def Redo(self):
        """Redoes transaction"""
        undo_operations = []
        while len(self._operations) > 0:
            undo_operations.append(self._operations.pop().Redo())
        self._operations = undo_operations
        return self

    def RefreshChanges(self):
        """Ensures visual update of saved states"""
        for operation in self._operations:
            if type(operation) is UndoChange:
                operation._obj.OnUndoRedoChanged()


class UndoNew(object):
    """Implements a undo new operation"""
    def __init__(self, obj):
        """Init"""
        obj.OnUndoRedoAdd()
        self._obj = obj

    def Dispose(self):
        """Remove associated data"""
        pass

    def Rollback(self):
        """Implements rollback"""
        self._obj.OnUndoRedoRemoving()
        self._obj.RemoveRelations()

    def Undo(self):
        """Implements undo"""
        return RedoNew(self._obj)


class UndoLoad(object):
    """Implements a undo load operation"""
    def __init__(self, obj):
        """Init"""
        obj.OnUndoRedoLoaded()
        self._obj = obj

    def Dispose(self):
        """Remove associated data"""
        pass

    def Rollback(self):
        """Implements rollback"""
        self._obj.OnUndoRedoUnloaded()
        self._obj.RemoveRelations()

    def Undo(self):
        """Implements undo"""
        return RedoLoad(self._obj)


class UndoUnload(object):
    """Implements a undo unload operation"""
    def __init__(self, obj):
        """Init"""
        obj.OnUndoRedoUnloaded()
        obj.RemoveRelations()
        self._obj = obj

    def Dispose(self):
        """Remove associated data"""
        pass

    def Rollback(self):
        """Implements rollback"""
        self._obj.RestoreRelations()
        self._obj.OnUndoRedoLoaded()

    def Undo(self):
        """Implements undo"""
        return RedoUnload(self._obj)


class RedoLoad(object):
    """Implements a redo for load operation"""
    def __init__(self, obj):
        """Initialize redo"""
        obj.OnUndoRedoUnloaded()
        obj.RemoveRelations()
        self._obj = obj

    def Dispose(self):
        """Remove associated data"""
        del self._obj  # unloaded objects dont interact

    def Redo(self):
        """Implements redo"""
        self._obj.RestoreRelations()
        return UndoLoad(self._obj)


class RedoUnload(object):
    """Implements a redo unload operation"""
    def __init__(self, obj):
        """Init"""
        obj.RestoreRelations()
        obj.OnUndoRedoLoaded()
        self._obj = obj

    def Dispose(self):
        """Remove associated data"""
        del self._obj  # unloaded objects dont interact
        pass

    def Undo(self):
        """Implements undo"""
        return UndoUnload(self._obj)


class RedoNew(object):
    """Implements a redo new operation"""
    def __init__(self, obj):
        """Initialize redo"""
        obj.OnUndoRedoRemoving()
        obj.RemoveRelations()
        self._obj = obj

    def Dispose(self):
        """Remove associated data"""
        self._obj.Dispose()

    def Redo(self):
        """Implements redo"""
        self._obj.RestoreRelations()
        return UndoNew(self._obj)


class UndoDelete(object):
    """Implements a undo delete operation"""
    def __init__(self, obj):
        """Initialize undo"""
        obj.OnUndoRedoRemoving()
        obj.RemoveRelations()
        self._obj = obj

    def Dispose(self):
        """Remove associated data"""
        self._obj.Dispose()

    def Rollback(self):
        """Implements rollback"""
        self._obj.RestoreRelations()
        self._obj.OnUndoRedoAdd()

    def Undo(self):
        """Implements undo"""
        return RedoDelete(self._obj)


class RedoDelete(object):
    """Implements a redo delete operation"""
    def __init__(self, obj):
        """Initialize redo"""
        obj.RestoreRelations()
        obj.OnUndoRedoAdd()
        self._obj = obj

    def Dispose(self):
        """Remove associated data"""
        pass

    def Redo(self):
        """Implements redo"""
        return UndoDelete(self._obj)


class UndoChange(object):
    """Implements a undo change operation"""
    def __init__(self, obj):
        """Initialize undo change"""
        self._oldObj = copy_temps(obj)
        self._obj = obj

    def Dispose(self):
        """Remove associated data"""
        self._oldObj.Dispose()

    def Rollback(self):
        """Implements rollback"""
        self._obj.RemoveRelations()
        self._obj .__dict__ = self._oldObj.__dict__.copy()
        self._obj.RestoreRelations()
        self._obj.OnUndoRedoChanged()

    def Undo(self):
        """Implements undo"""
        return RedoChange(self._oldObj, self._obj)


class RedoChange(object):
    """Implements a redo change operation"""
    def __init__(self, oldObj, obj):
        """Initialize redo change"""
        self._newObj = copy_temps(obj)
        self._obj = obj
        self._obj.RemoveRelations()
        self._obj .__dict__ = oldObj.__dict__.copy()
        respect_visuals_from(self._newObj, self._obj)
        self._obj.RestoreRelations()
        self._obj.OnUndoRedoChanged()

    def Dispose(self):
        """Remove associated data"""
        self._newObj.Dispose()

    def Redo(self):
        """Implements redo"""
        r = UndoChange(self._obj)
        self._obj.RemoveRelations()
        self._obj .__dict__ = self._newObj.__dict__.copy()
        self._obj.RestoreRelations()
        self._obj.OnUndoRedoChanged()
        return r


class TransactionStack(object):
    """Implements a queue of stacked operations"""
    instance = None
    delayedCalls = {}
    delayedCallsFiltered = []

    @staticmethod
    def DoBeginTransaction(name=None):
        """Starts a new transaction"""
        TransactionStack.instance.BeginTransaction(name)

    @staticmethod
    def DoCommit():
        """Commit transaction"""
        TransactionStack.instance.Commit()

    @staticmethod
    def DoRollback():
        """Rollback transaction"""
        TransactionStack.instance.Rollback()

    @staticmethod
    def DoUndo():
        """Rollback transaction"""
        TransactionStack.instance.Undo()

    @staticmethod
    def DoRedo():
        """Rollback transaction"""
        TransactionStack.instance.Redo()

    @staticmethod
    def DoCreate(obj):
        """Create transaction"""
        TransactionStack.instance.Create(obj)

    @staticmethod
    def DoLoad(obj):
        """Load transaction"""
        TransactionStack.instance.Load(obj)

    @staticmethod
    def DoUnload(obj):
        """Unload transaction"""
        TransactionStack.instance.Unload(obj)

    @staticmethod
    def DoDelete(obj):
        """Create transaction"""
        TransactionStack.instance.Delete(obj)

    @staticmethod
    def DoSaveState(obj):
        """Save object state"""
        TransactionStack.instance.SaveState(obj)

    @staticmethod
    def InTransaction():
        """Returns info about undo/redo"""
        return TransactionStack.instance._tran is not None

    @staticmethod
    def InUndoRedo():
        """Returns info about undo/redo"""
        return TransactionStack.instance._inUndoRedo

    @staticmethod
    def InUndo():
        """Returns info about undo/redo"""
        return TransactionStack.instance._inUndo

    @staticmethod
    def InRedo():
        """Returns info about undo/redo"""
        return TransactionStack.instance._inRedo

    @staticmethod
    def DoSetKey(key):
        """Set the key for the current, uncommited transaction"""
        return TransactionStack.instance.SetKey(key)

    def __init__(self, deep):
        """Initialize the transaction stack"""
        self._deep = deep
        self._undo = []
        self._redo = []
        self._tran = None
        self._inUndoRedo = False
        self._inRedo = False
        self._inUndo = False
        self._logger = None
        TransactionStack.instance = self

    def TransactionStatus(self, hashcode):
        """This method is a facility for loggers to see the
        current status of some log entry. The return code
        will be:
            1 ... transaction exists in the redo stack
            0 ... transaction exists in the undo stack
           -1 ... transaction don't exist anymore (may be deleted from some stack)'"""
        if hashcode in (hash(tran) for tran in self._redo):
            return 1
        if hashcode in (hash(tran) for tran in self._undo):
            return 0
        return -1

    def SetLogger(self, logger):
        """Set an method logger that will be called for:

            a) Each new transaction commit
            b) Each new transaction undo
            c) Each new transaction redo

            The logger will receive the transaction
            and a text mode 'commit', 'undo' or 'redo'.
            Indeed, the logger will receive 'start'
            and 'end' messages (without transaction)
            for the SetLogger and for SetLogger(None).
            logger = fn(command, transaction=None)

            The obvious mission : create logs for projects
            and/or workspace
        """
        if self._logger == logger:
            # stupid call will be dissmissed
            return
        # finalize previous loggers
        if self._logger:
            self._logger('end')

        self._logger = logger
        # start new transaction logger
        if self._logger:
            self._logger('start')

    def BeginTransaction(self, name=''):
        """Starts a new transaction"""
        if self._tran is None:
            self._tran = Transaction(name)
            self._inTran = True
        return self._tran

    def Commit(self):
        """Commit an open transaction"""
        if self._tran is None:
            return
        self._inUndoRedo = False
        self._inRedo = False
        self._inUndo = False
        self._tran.RefreshChanges()
        self._undo.append(self._tran)
        if self._logger:
            self._logger('commit', transaction=self._tran)
        self._tran = None
        self.CleanRedo()
        for method in TransactionStack.delayedCalls:
            if method in TransactionStack.delayedCallsFiltered:
                continue
            for l in TransactionStack.delayedCalls[method]:
                method(*l[0], **l[1])
        TransactionStack.delayedCalls = {}
        TransactionStack.delayedCallsFiltered = []
        while len(self._undo) > self._deep:
            self._undo[0].Dispose()
            del self._undo[0]

    def Rollback(self):
        """Rollback an open transaction"""
        if self._tran is None:
            return
        self._inUndoRedo = True
        self._tran.Rollback()
        TransactionStack.delayedCalls = {}
        TransactionStack.delayedCallsFiltered = []
        self._tran = None
        self._inUndoRedo = False

    def Undo(self):
        """Undoes the last operation"""
        self._inUndoRedo = True
        self._inUndo = True
        if not self._tran is None:
            self.Rollback()
        elif len(self._undo) > 0:
            if self._logger:
                self._logger('undo')
            self._redo.append(self._undo.pop().Undo())
        self._inUndoRedo = False
        self._inUndo = False
        while TransactionStack.delayedCalls:
            k = TransactionStack.delayedCalls[0]
            k[0](*k[1], **k[2])
            del TransactionStack.delayedCalls[0]
        TransactionStack.delayedCallsFiltered = []

    def Redo(self):
        """Redoes the last operation"""
        self._inUndoRedo = True
        self._inRedo = True
        if not self._tran is None:
            self.Rollback()
        if len(self._redo) > 0:
            if self._logger:
                self._logger('redo')
            self._undo.append(self._redo.pop().Redo())
        self._inUndoRedo = False
        self._inRedo = False
        while TransactionStack.delayedCalls:
            k = TransactionStack.delayedCalls[0]
            k[0](*k[1], **k[2])
            del TransactionStack.delayedCalls[0]
        TransactionStack.delayedCallsFiltered = []

    def CanRedo(self):
        """Check about redo availability"""
        return len(self._redo) > 0

    def CanUndo(self):
        """Check about undo availability"""
        return len(self._undo) > 0

    def CleanRedo(self):
        """Clean redo stack"""
        for k in self._redo:
            k.Dispose()
        self._redo = []

    def CleanUndo(self):
        """Clean undo stack"""
        for k in self._undo:
            k.Dispose()
        self._undo = []

    def UndoName(self):
        """Get undo name"""
        if len(self._undo) > 0:
            return self._undo[-1]._name
        return None

    def SetKey(self, key):
        """Sets the key for the current, uncommited transaction"""
        if self._tran:
            self._tran.SetKey(key)
            return True
        return False

    def GetName(self):
        """Get the current, uncommited transaction name"""
        if self._tran:
            return self._tran._name
        return None

    def SetName(self, name):
        """Set the current, uncommited transaction name"""
        if self._tran:
            self._tran._name = name
        return None

    def RedoName(self):
        """Get redo name"""
        if len(self._redo) > 0:
            return self._redo[-1]._name

    def SaveState(self, obj):
        """Save object state"""
        if self._tran is None:
            self.BeginTransaction()
        if not self.SavedState(obj):
            # some operations can trigger others so we must ensure order
            pos = len(self._tran._operations)
            self._tran._operations.insert(pos, UndoChange(obj))

    def SavedState(self, obj):
        """Checks about saved state"""
        if self._tran is None:
            return False
        for operation in self._tran._operations:
            if type(operation) is UndoChange:
                if operation._obj == obj:
                    return True
        return False

    def Delete(self, obj):
        """Delete object"""
        if self._tran is None:
            self.BeginTransaction()
        # some operations can trigger others so we must ensure order
        pos = len(self._tran._operations)
        self._tran._operations.insert(pos, UndoDelete(obj))

    def Create(self, obj):
        """Create object"""
        if self._tran is None:
            self.BeginTransaction()
            try:
                pos = len(self._tran._operations)
                self._tran._operations.insert(pos, UndoNew(obj))
                self.Commit()
            except:
                self.Rollback()
        else:
            # some operations can trigger others so we must ensure order
            pos = len(self._tran._operations)
            self._tran._operations.insert(pos, UndoNew(obj))

    def Load(self, obj):
        """Load object"""
        if self._tran is None:
            self.BeginTransaction()
        # some operations can trigger others so we must ensure order
        pos = len(self._tran._operations)
        self._tran._operations.insert(pos, UndoLoad(obj))

    def Unload(self, obj):
        """Unload object"""
        if self._tran is None:
            self.BeginTransaction()
        # some operations can trigger others so we must ensure order
        pos = len(self._tran._operations)
        self._tran._operations.insert(pos, UndoUnload(obj))




"""Subclass of NewMember, which is generated by wxFormBuilder."""

import copy
import wx
import wxx
from activity.models.ui import ui as ui
import model
import app.resources as rc


# Implementing NewMember
class MemberDialog(ui.NewMember):
    """
    This dialog allows to setup data member of class
    or struct. You can set default value for using
    in constructors or as initialization of static
    members.
    """
    @wxx.SetInfo(__doc__)
    def __init__(self, parent, container):
        """Dialog initialization"""
        super(MemberDialog, self).__init__(parent)
        self._container = container
        scoped = lambda x: (hasattr(x, 'scoped') and x.scoped) or x.name
        self._types = dict([scoped(x), x] for x in container.types)
        self._autoname = ''  # proposed name
        # add types but not on-the-fly template type
        self.m_type.AppendItems([x for x in self._types.keys() if x != '@'])
        # we need to add types from template nested classes
        classes = container.nested_classes
        self._nested_template_types = []
        for clase in classes:
            for x in clase._template_types:
                if x not in self._nested_template_types:
                    self._nested_template_types.append(scoped(x))
        if len(self._nested_template_types) > 0:
            self.m_type.AppendItems(self._nested_template_types)
        self.choiceStr = ""
        self.m_type.SetFocus()
        icon = wx.EmptyIcon()
        icon.CopyFromBitmap(rc.GetBitmap("member"))
        self.SetIcon(icon)
        self._register_keybindings()

    def AutoName(self):
        """Suggest the argument name, based on type"""
        iSel = self.m_type.GetCurrentSelection()
        if iSel == wx.NOT_FOUND:
            return
        s = self.m_name.GetValue()
        if self._autoname != s and s:
            return
        kwargs = {
            'const': (self.m_const.IsChecked() and 'c') or '',
            'reference': (self.m_reference.IsChecked() and 'r') or '',
            'ptr': (self.m_ptr.IsChecked() and 'p') or '',
            'pptr': (self.m_pptr.IsChecked() and 'p') or '',
            'constptr': (self.m_constptr.IsChecked() and 'c') or '',
            'array': (self.m_array.IsChecked() and 'a') or '',
            'typename': self.m_type.GetString(iSel).replace('::', '_'),
        }
        #volatile = (self.m_volatile.IsChecked() and 'v') or ''
        self._autoname = '{const}{reference}{ptr}{pptr}{constptr}{array}{typename}'.format(
            **kwargs)
        self.m_name.SetValue(self._autoname)

    def _register_keybindings(self):
        """Register accelerators for static labels that must change the focus"""
        newId_t = wx.NewId()
        newId_n = wx.NewId()
        newId_a = wx.NewId()
        newId_d = wx.NewId()
        newId_o = wx.NewId()
        self.Bind(wx.EVT_MENU, self.OnActivateType, id=newId_t)
        self.Bind(wx.EVT_MENU, self.OnActivateName, id=newId_n)
        self.Bind(wx.EVT_MENU, self.OnActivateAccess, id=newId_a)
        self.Bind(wx.EVT_MENU, self.OnActivateDefault, id=newId_d)
        self.Bind(wx.EVT_MENU, self.OnActivateNotes, id=newId_o)
        aTable = wx.AcceleratorTable([
            wx.AcceleratorEntry(wx.ACCEL_ALT, ord('T'), newId_t),
            wx.AcceleratorEntry(wx.ACCEL_ALT, ord('N'), newId_n),
            wx.AcceleratorEntry(wx.ACCEL_ALT, ord('A'), newId_a),
            wx.AcceleratorEntry(wx.ACCEL_ALT, ord('D'), newId_d),
            wx.AcceleratorEntry(wx.ACCEL_ALT, ord('O'), newId_o)
            ])
        self.SetAcceleratorTable(aTable)

    def OnActivateType(self, event):
        """activate type combo"""
        self.m_type.SetFocus()

    def OnActivateName(self, event):
        """activate name entry"""
        self.m_name.SetFocus()

    def OnActivateAccess(self, event):
        """activate acces combo"""
        self.m_choice2.SetFocus()

    def OnActivateDefault(self, event):
        """activate default value"""
        self.m_textCtrl8.SetFocus()

    def OnActivateNotes(self, event):
        """Activate notes"""
        self.m_richText1.SetFocus()

    def OnEnterName(self, event):
        """This event is generated when the enter is pressed in the
        name entry"""
        self.m_choice2.SetFocus()

    def OnTypeChanged(self, event):
        """This event happens when the return type is changed. The main goal
        of this callback is handling template types for argument specification"""
        iSel = self.m_type.GetCurrentSelection()
        _type = self._types.get(self.m_type.GetString(iSel), None)
        template_args = False
        if _type is not None:
            if _type._template is not None:
                template_args = True
        if template_args is True:
            self.m_staticText67.Enable(True)
            self.m_template_args.Enable(True)
            self.m_staticText68.Enable(True)
        else:
            self.m_staticText67.Enable(False)
            self.m_template_args.Enable(False)
            self.m_staticText68.Enable(False)
            self.m_template_args.SetValue('')
        self.AutoName()

    def CopyAttributes(self, member):
        """Get the atributes"""
        member._name = self._name
        member._typei = copy.copy(self._typei)
        member._access = self._access
        member._static = self._static
        member._default = self._default
        member._volatile = self._volatile
        member._mutable = self._mutable
        member._bitField = self._bitField
        if self._bitField:
            member._bitFieldSize = self._bitFieldSize
        member._note = self._note
        member.inner_class.AutoInit()

    def SetAttributes(self, member):
        """Set the attributes"""
        self.m_name.SetValue(member._name)
        ti = member._typei
        iSel = self.m_type.FindString(ti.scoped)
        self.m_type.SetSelection(iSel)
        iSel = self.m_choice2.FindString(member._access)
        self.m_choice2.SetSelection(iSel)
        self.m_checkBox105.SetValue(member._static)
        self.m_textCtrl8.SetValue(member._default)
        self.m_checkBox49.SetValue(member._volatile)
        self.m_checkBox48.SetValue(member._mutable)
        self.m_const.SetValue(ti._const)
        self.m_ptr.SetValue(ti._ptr)
        self.m_reference.SetValue(ti._ref)
        self.m_pptr.SetValue(ti._ptr_to_ptr)
        self.m_constptr.SetValue(ti._const_ptr)
        self.m_array.SetValue(ti._array)
        if ti._array is True:
            self.m_textCtrl7.Show(True)
            self.m_textCtrl7.Enable(True)
            self.m_textCtrl7.SetValue(str(ti._array_size))
        else:
            self.m_textCtrl7.SetValue('0')
        self.m_checkBox51.SetValue(member._bitField)
        if ti._type_args is not None:
            self.m_staticText67.Enable(True)
            self.m_template_args.Enable(True)
            self.m_staticText68.Enable(True)
            self.m_template_args.SetValue(ti._type_args)

        if member._bitField is True:
            self.m_textCtrl39.Show(True)
            self.m_textCtrl39.Enable(True)
            self.m_textCtrl39.SetValue(str(member._bitFieldSize))
        self.m_richText1.SetValue(member._note)
        self.SetTitle("Edit member")

    def Validate(self):
        """Dialog validation"""
        self._name = self.m_name.GetValue()
        if len(self._name) == 0:
            wx.MessageBox("Member name must not be empty", "Error",
                wx.OK | wx.CENTER | wx.ICON_ERROR, self)
            return False
        iSel = self.m_type.GetCurrentSelection()
        if iSel == wx.NOT_FOUND:
            wx.MessageBox("Invalid type", "Error",
                wx.OK | wx.CENTER | wx.ICON_ERROR, self)
            return False
        typename = self.m_type.GetString(iSel)
        iSel = self.m_choice2.GetCurrentSelection()
        if iSel == wx.NOT_FOUND:
            wx.MessageBox("Invalid access", "Error",
                wx.OK | wx.CENTER | wx.ICON_ERROR, self)
            return False
        self._static = self.m_checkBox105.IsChecked()
        self._access = self.m_choice2.GetString(iSel)
        self._default = self.m_textCtrl8.GetValue()
        self._volatile = self.m_checkBox49.GetValue()
        self._mutable = self.m_checkBox48.GetValue()
        if self.m_array.IsChecked():
            try:
                asize = int(self.m_textCtrl7.GetValue())
            except:
                asize = ''
        else:
            asize = None
        if typename in self._nested_template_types:
            self._typei = model.cc.typeinst(
                type=self._types['@'],
                type_alias=typename,
                const=self.m_const.IsChecked(),
                ptr=self.m_ptr.IsChecked(),
                ref=self.m_reference.IsChecked(),
                ptrptr=self.m_pptr.IsChecked(),
                constptr=self.m_constptr.IsChecked(),
                array=self.m_array.IsChecked(),
                arraysize=asize
                )
        else:
            _type = self._types[typename]
            if _type._template is not None:
                #we construct type instance with explicit arguments
                type_args = self.m_template_args.GetValue()
                self._typei = model.cc.typeinst(
                    type=_type,
                    type_args=type_args,
                    const=self.m_const.IsChecked(),
                    ptr=self.m_ptr.IsChecked(),
                    ref=self.m_reference.IsChecked(),
                    ptrptr=self.m_pptr.IsChecked(),
                    constptr=self.m_constptr.IsChecked(),
                    array=self.m_array.IsChecked(),
                    arraysize=asize
                )
            else:
                self._typei = model.cc.typeinst(
                    type=self._types[typename],
                    const=self.m_const.IsChecked(),
                    ptr=self.m_ptr.IsChecked(),
                    ref=self.m_reference.IsChecked(),
                    ptrptr=self.m_pptr.IsChecked(),
                    constptr=self.m_constptr.IsChecked(),
                    array=self.m_array.IsChecked(),
                    arraysize=asize
            )
        self._bitField = self.m_checkBox51.IsChecked()
        if self._bitField is True:
            self._bitFieldSize = int(self.m_textCtrl39.GetValue())
        else:
            self._bitFieldSize = 0
        self._note = self.m_richText1.GetValue()
        return True

    def get_kwargs(self):
        """return arguments for object instance"""
        return  {'parent': self._container, 'name': self._name,
            'type': self._typei, 'access': self._access, 'static': self._static,
            'volatile': self._volatile, 'mutable': self._mutable,
            'bitfield': self._bitField, 'bitfieldsize': self._bitFieldSize,
            'default': self._default}

    # Handlers for NewMember events.
    def OnKeyDown(self, event):
        """Listbox selection"""
        keycode = event.GetKeyCode()
        if keycode == wx.WXK_UP or keycode == wx.WXK_NUMPAD_UP:
            i = self.m_type.GetSelection()
            if i is not wx.NOT_FOUND and i > 0:
                self.m_type.SetSelection(i - 1)
        elif keycode == wx.WXK_DOWN or keycode == wx.WXK_NUMPAD_DOWN:
            i = self.m_type.GetSelection() + 1
            if i > wx.NOT_FOUND and i < len(self._types):
                self.m_type.SetSelection(i)
        elif keycode < 256:
            keychar = chr(keycode)
            if keychar.isalnum() or keycode is wx.WXK_SPACE:
                self.choiceStr += keychar.lower()
                for t in self._types:
                    tl = t.lower()
                    if tl.find(self.choiceStr) == 0:
                        sel = self.m_type.FindString(t)
                        if sel is not wx.NOT_FOUND:
                            self.m_type.SetSelection(sel)
                            if keycode is not wx.WXK_SPACE:
                                event.Skip()
                            return
            self.choiceStr = ""
        event.Skip()

    def OnPointerToggle(self, event):
        """ptr toggle gui"""
        if self.m_ptr.IsChecked():
            self.m_constptr.Enable(True)
            self.m_checkBox50.Enable(True)
            self.m_pptr.Enable(True)
        else:
            self.m_constptr.Enable(False)
            self.m_checkBox50.Enable(False)
            self.m_pptr.Enable(False)
            self.m_constptr.SetValue(False)
            self.m_checkBox50.SetValue(False)
            self.m_pptr.SetValue(False)
        self.AutoName()

    def OnToggleArray(self, event):
        "toggle array event"
        if  self.m_array.IsChecked():
            self.m_checkBox51.SetValue(False)
            self.m_textCtrl39.Show(False)
            self.m_textCtrl39.Enable(False)
            self.m_textCtrl7.Show(True)
            self.m_textCtrl7.Enable(True)
        else:
            self.m_textCtrl7.Show(False)
            self.m_textCtrl7.Enable(False)
        self.AutoName()

    def OnToggleStatic(self, event):
        """toggle static event"""
        if self.m_checkBox105.IsChecked():
            #disable bit field
            self.m_checkBox51.SetValue(False)
            self.m_checkBox51.Enable(False)
            self.m_textCtrl39.Show(False)
            self.m_textCtrl39.Enable(False)
            #disable mutable
            self.m_checkBox48.SetValue(False)
            self.m_checkBox48.Enable(False)
        else:
            self.m_checkBox51.Enable(True)
            self.m_checkBox48.Enable(True)
        event.Skip()

    def OnToggleBitFiled(self, event):
        "toggle array event"
        if  self.m_checkBox51.IsChecked():
            self.m_array.SetValue(False)
            self.m_textCtrl7.Show(False)
            self.m_textCtrl7.Enable(False)
            self.m_textCtrl39.Show(True)
            self.m_textCtrl39.Enable(True)
        else:
            self.m_textCtrl39.Show(False)
            self.m_textCtrl39.Enable(False)
        event.Skip()

    def OnCancel(self, event):
        """cancel event handler"""
        self.EndModal(wx.ID_CANCEL)

    def OnOK(self, event):
        """ok event handler"""
        if self.Validate():
            self.EndModal(wx.ID_OK)


